"""
Worker Manager for handling various application workers.

This class uses inline imports to avoid slow startup times. Do not move
imports to the top of the file.
"""

from typing import Dict
from airunner.components.application.workers.worker import Worker
from airunner.enums import SignalCode
from airunner.utils.application.create_worker import create_worker


class WorkerManager(Worker):
    def __init__(self, *args, **kwargs):
        self.signal_handlers = {
            SignalCode.LLM_TEXT_GENERATE_REQUEST_SIGNAL: self.on_llm_request_signal,
            SignalCode.START_AUTO_IMAGE_GENERATION_SIGNAL: self.on_start_auto_image_generation_signal,
            SignalCode.DO_GENERATE_SIGNAL: self.on_do_generate_signal,
            SignalCode.TTS_GENERATOR_WORKER_ADD_TO_STREAM_SIGNAL: self.on_tts_generator_worker_add_to_stream_signal,
            SignalCode.TTS_ENABLE_SIGNAL: self.on_enable_tts_signal,
            SignalCode.STT_LOAD_SIGNAL: self.on_stt_load_signal,
            SignalCode.STT_START_CAPTURE_SIGNAL: self.on_stt_start_capture_signal,
            SignalCode.ART_MODEL_DOWNLOAD_REQUIRED: self.on_art_model_download_required,
            SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE: self.on_huggingface_download_complete,
            SignalCode.SD_UNLOAD_SIGNAL: self.on_unload_art_signal,
            SignalCode.SD_CANCEL_SIGNAL: self.on_sd_cancel_signal,
            SignalCode.STOP_AUTO_IMAGE_GENERATION_SIGNAL: self.on_stop_auto_image_generation_signal,
            SignalCode.INTERRUPT_IMAGE_GENERATION_SIGNAL: self.on_interrupt_image_generation_signal,
            SignalCode.CHANGE_SCHEDULER_SIGNAL: self.on_change_scheduler_signal,
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL: self.on_model_status_changed_signal,
            SignalCode.SD_LOAD_SIGNAL: self.on_load_art_signal,
            SignalCode.SD_ART_MODEL_CHANGED: self.on_art_model_changed,
            SignalCode.CONTROLNET_LOAD_SIGNAL: self.on_load_controlnet_signal,
            SignalCode.CONTROLNET_UNLOAD_SIGNAL: self.on_unload_controlnet_signal,
            SignalCode.SAFETY_CHECKER_LOAD_SIGNAL: self.on_safety_checker_load_signal,
            SignalCode.SAFETY_CHECKER_UNLOAD_SIGNAL: self.on_safety_checker_unload_signal,
            SignalCode.INPUT_IMAGE_SETTINGS_CHANGED: self.on_input_image_settings_changed_signal,
            SignalCode.LORA_UPDATE_SIGNAL: self.on_update_lora_signal,
            SignalCode.EMBEDDING_UPDATE_SIGNAL: self.on_update_embeddings_signal,
            SignalCode.EMBEDDING_DELETE_MISSING_SIGNAL: self.delete_missing_embeddings,
            SignalCode.LLM_UNLOAD_SIGNAL: self.on_llm_on_unload_signal,
            SignalCode.LLM_LOAD_SIGNAL: self.on_llm_load_model_signal,
            SignalCode.LLM_MODEL_CHANGED: self.on_llm_model_changed_signal,
            SignalCode.LLM_MODEL_DOWNLOAD_REQUIRED: self.on_llm_model_download_required_signal,
            SignalCode.RAG_RELOAD_INDEX_SIGNAL: self.on_llm_reload_rag_index_signal,
            SignalCode.RAG_INDEX_ALL_DOCUMENTS: self.on_rag_index_all_documents_signal,
            SignalCode.RAG_INDEX_SELECTED_DOCUMENTS: self.on_rag_index_selected_documents_signal,
            SignalCode.RAG_INDEX_CANCEL: self.on_rag_index_cancel_signal,
            SignalCode.RAG_LOAD_DOCUMENTS: self.on_rag_load_documents_signal,
            SignalCode.INDEX_DOCUMENT: self.on_index_document_signal,
            SignalCode.LLM_START_FINE_TUNE: self.on_llm_start_fine_tune_signal,
            SignalCode.LLM_FINE_TUNE_CANCEL: self.on_llm_fine_tune_cancel_signal,
            SignalCode.LLM_START_QUANTIZATION: self.on_llm_start_quantization_signal,
            SignalCode.LLM_CLEAR_HISTORY_SIGNAL: self.on_llm_clear_history_signal,
            SignalCode.ADD_CHATBOT_MESSAGE_SIGNAL: self.on_llm_add_chatbot_response_to_history,
            SignalCode.LOAD_CONVERSATION: self.on_llm_load_conversation,
            SignalCode.INTERRUPT_PROCESS_SIGNAL: self.on_interrupt_process_signal,
            SignalCode.QUIT_APPLICATION: self.on_quit_application_signal,
            SignalCode.CONVERSATION_DELETED: self.on_conversation_deleted_signal,
            SignalCode.SECTION_CHANGED: self.on_section_changed_signal,
            SignalCode.GENERATE_MASK: self.on_generate_mask_signal,
            SignalCode.AUDIO_CAPTURE_WORKER_RESPONSE_SIGNAL: self.on_stt_process_audio_signal,
            SignalCode.STT_STOP_CAPTURE_SIGNAL: self.on_stt_stop_capture_signal,
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL: self.on_model_status_changed_signal,
            SignalCode.RECORDING_DEVICE_CHANGED: self.on_recording_device_changed_signal,
            SignalCode.STT_UNLOAD_SIGNAL: self.on_stt_unload_signal,
            SignalCode.UNBLOCK_TTS_GENERATOR_SIGNAL: self.on_unblock_tts_generator_signal,
            SignalCode.TTS_DISABLE_SIGNAL: self.on_disable_tts_signal,
            SignalCode.LLM_TEXT_STREAMED_SIGNAL: self.on_llm_text_streamed_signal,
            SignalCode.TTS_MODEL_CHANGED: self._reload_tts_model_manager,
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL: self.on_application_settings_changed_signal,
            SignalCode.TTS_QUEUE_SIGNAL: self.on_add_to_queue_signal,
            SignalCode.PLAYBACK_DEVICE_CHANGED: self.on_playback_device_changed_signal,
            SignalCode.IMAGE_EXPORTED: self.on_image_exported_signal,
            SignalCode.FARA_LOAD_SIGNAL: self.on_fara_load_signal,
            SignalCode.FARA_UNLOAD_SIGNAL: self.on_fara_unload_signal,
            SignalCode.FARA_MODEL_DOWNLOAD_REQUIRED: self.on_fara_model_download_required_signal,
            SignalCode.START_HUGGINGFACE_DOWNLOAD: self.on_start_huggingface_download_signal,
            SignalCode.START_OPENVOICE_BATCH_DOWNLOAD: self.on_start_openvoice_batch_download_signal,
        }
        super().__init__()
        self._mask_generator_worker = None
        self._sd_worker = None
        self._safety_checker_worker = None
        self._fara_worker = None
        self._pending_generation_request = None
        self._download_dialog = (
            None  # Store dialog reference to prevent garbage collection
        )
        self._stt_audio_capture_worker = None
        self._stt_audio_processor_worker = None
        self._tts_generator_worker = None
        self._tts_vocalizer_worker = None
        self._llm_generate_worker = None
        self._document_worker = None
        self._huggingface_download_worker = None
        self._image_export_worker = None
        self._model_scanner_worker = None
        if self.logger:
            self.logger.debug(
                f"WorkerManager initialized. Mediator ID: {id(self.mediator)}"
            )

        self.model_scanner_worker.add_to_queue("scan_for_models")

    def handle_message(self, message: Dict):
        if self.logger:
            self.logger.info(
                f"WorkerManager::handle_message CALLED with request_type={message.get('request_type')}"
            )
        data = message.get("data", {})
        request_type = message.get("request_type")
        try:
            if request_type == "llm_generate":
                if self.llm_generate_worker is not None:
                    self.llm_generate_worker.on_llm_request_signal(data)
            elif request_type == "fara_generate":
                # Route USE_COMPUTER requests to Fara worker
                if self.fara_worker is not None:
                    self.fara_worker.on_fara_request_signal(data)
            elif request_type == "image_auto_generate":
                if self.sd_worker is not None:
                    self.sd_worker.on_start_auto_image_generation_signal(data)
            elif request_type == "image_generate":
                # Intercept image generation to ensure safety checker is ready
                self._handle_image_generation_request(data)
            elif request_type == "tts_generate":
                if self.tts_vocalizer_worker is not None:
                    self.tts_vocalizer_worker.on_tts_generator_worker_add_to_stream_signal(
                        data
                    )
            elif request_type == "tts_enable":
                if self.tts_generator_worker is not None:
                    self.tts_generator_worker.on_enable_tts_signal(data)
            elif request_type == "stt_load":
                if self.stt_audio_processor_worker is not None:
                    self.stt_audio_processor_worker.on_stt_load_signal(data)
            elif request_type == "stt_start_capture":
                if self.stt_audio_capture_worker is not None:
                    self.stt_audio_capture_worker.on_stt_start_capture_signal(
                        data
                    )

        except Exception as e:
            if self.logger:
                self.logger.error(f"Error processing worker requests: {e}")

    @property
    def model_scanner_worker(self):
        if self._model_scanner_worker is None:
            from airunner.components.application.workers.model_scanner_worker import (
                ModelScannerWorker,
            )

            self._model_scanner_worker = create_worker(ModelScannerWorker)
        return self._model_scanner_worker

    @property
    def image_export_worker(self):
        if self._image_export_worker is None:
            from airunner.components.art.workers.image_export_worker import (
                ImageExportWorker,
            )

            self._image_export_worker = create_worker(ImageExportWorker)
        return self._image_export_worker

    @property
    def safety_checker_worker(self):
        if self._safety_checker_worker is None:
            from airunner.components.art.workers.safety_checker_worker import (
                SafetyCheckerWorker,
            )

            self._safety_checker_worker = create_worker(SafetyCheckerWorker)
        return self._safety_checker_worker

    @property
    def mask_generator_worker(self):
        if self._mask_generator_worker is None:
            from airunner.components.art.workers.mask_generator_worker import (
                MaskGeneratorWorker,
            )

            self._mask_generator_worker = create_worker(MaskGeneratorWorker)
        return self._mask_generator_worker

    @property
    def sd_worker(self):
        if self._sd_worker is None:
            from airunner.components.art.workers.sd_worker import SDWorker

            self._sd_worker = create_worker(
                SDWorker, image_export_worker=self.image_export_worker
            )
            # reference self.mask_generator_worker to ensure it is created
            _ = self.mask_generator_worker
        return self._sd_worker

    @property
    def stt_audio_capture_worker(self):
        if self._stt_audio_capture_worker is None:
            from airunner.components.stt.workers.audio_capture_worker import (
                AudioCaptureWorker,
            )

            self._stt_audio_capture_worker = create_worker(AudioCaptureWorker)
        return self._stt_audio_capture_worker

    @property
    def stt_audio_processor_worker(self):
        if self._stt_audio_processor_worker is None:
            from airunner.components.stt.workers.audio_processor_worker import (
                AudioProcessorWorker,
            )

            self._stt_audio_processor_worker = create_worker(
                AudioProcessorWorker
            )
        return self._stt_audio_processor_worker

    @property
    def tts_generator_worker(self):
        if self._tts_generator_worker is None:
            from airunner.components.tts.workers.tts_generator_worker import (
                TTSGeneratorWorker,
            )

            self._tts_generator_worker = create_worker(TTSGeneratorWorker)
        return self._tts_generator_worker

    @property
    def tts_vocalizer_worker(self):
        if self._tts_vocalizer_worker is None:
            from airunner.components.tts.workers.tts_vocalizer_worker import (
                TTSVocalizerWorker,
            )

            self._tts_vocalizer_worker = create_worker(TTSVocalizerWorker)
        return self._tts_vocalizer_worker

    @property
    def llm_generate_worker(self):
        if self._llm_generate_worker is None:
            from airunner.components.llm.workers.llm_generate_worker import (
                LLMGenerateWorker,
            )

            self._llm_generate_worker = create_worker(LLMGenerateWorker)
        return self._llm_generate_worker

    @property
    def fara_worker(self):
        if self._fara_worker is None:
            from airunner.components.llm.workers.fara_worker import (
                FaraWorker,
            )

            self._fara_worker = create_worker(FaraWorker)
        return self._fara_worker

    @property
    def document_worker(self):
        if self._document_worker is None:
            from airunner.components.documents.workers.document_worker import (
                DocumentWorker,
            )

            self._document_worker = create_worker(DocumentWorker)
        return self._document_worker

    @property
    def huggingface_download_worker(self):
        if self._huggingface_download_worker is None:
            from airunner.components.application.workers.huggingface_download_worker import (
                HuggingFaceDownloadWorker,
            )

            self._huggingface_download_worker = create_worker(
                HuggingFaceDownloadWorker
            )
        return self._huggingface_download_worker

    def on_llm_request_signal(self, data: Dict):
        if self.logger:
            self.logger.info(
                f"WorkerManager::on_llm_request_signal CALLED with data keys: {list(data.keys())}"
            )

        # Check if this is a USE_COMPUTER action - route to Fara worker
        from airunner.enums import LLMActionType

        request_data = data.get("request_data", {})
        action = request_data.get("action")

        # Handle both enum and string action values
        is_use_computer = False
        if action is LLMActionType.USE_COMPUTER:
            is_use_computer = True
        elif isinstance(action, str) and action.lower() == "use_computer":
            is_use_computer = True

        if is_use_computer:
            if self.logger:
                self.logger.info("USE_COMPUTER action detected - routing to Fara worker")
            # First unload all other models, then route to Fara
            self.emit_signal(SignalCode.FARA_LOAD_SIGNAL, {})
            self.add_to_queue(
                {
                    "data": data,
                    "request_type": "fara_generate",
                }
            )
        else:
            self.add_to_queue(
                {
                    "data": data,
                    "request_type": "llm_generate",
                }
            )

    def on_start_auto_image_generation_signal(self, data: Dict):
        self.add_to_queue(
            {
                "data": data,
                "request_type": "image_auto_generate",
            }
        )

    def on_do_generate_signal(self, data: Dict):
        self.add_to_queue(
            {
                "data": data,
                "request_type": "image_generate",
            }
        )

    def on_tts_generator_worker_add_to_stream_signal(self, data: Dict):
        self.add_to_queue({"data": data, "request_type": "tts_generate"})

    def on_enable_tts_signal(self, data: Dict):
        self.add_to_queue(
            {
                "data": data,
                "request_type": "tts_enable",
            }
        )

    def on_stt_load_signal(self, data: Dict):
        self.add_to_queue(
            {
                "data": data,
                "request_type": "stt_load",
            }
        )

    def on_stt_start_capture_signal(self, data: Dict):
        self.add_to_queue(
            {
                "data": data,
                "request_type": "stt_start_capture",
            }
        )

    def on_art_model_download_required(self, data: Dict):
        """Handle art model download requirement.

        Args:
            data: Download info with repo_id, model_path, missing_files, version, etc.
        """
        import os
        from airunner.components.llm.gui.windows.huggingface_download_dialog import (
            HuggingFaceDownloadDialog,
        )

        repo_id = data.get("repo_id")
        model_path = data.get("model_path")
        missing_files = data.get("missing_files", [])
        version = data.get("version", "")
        pipeline_action = data.get("pipeline_action", "txt2img")

        # Store the pending image generation request so we can retry after download
        image_request = data.get("image_request")
        if image_request:
            self._pending_generation_request = (
                data  # Store the full data dict, not just image_request
            )
            if self.logger:
                self.logger.info(
                    "Stored pending generation request to retry after download"
                )

        if self.logger:
            self.logger.info(
                f"WorkerManager: Starting download for {repo_id} ({len(missing_files)} missing files)"
            )

        # Determine model type for download worker
        # All art models (SDXL, FLUX, SD) use "art" type for bootstrap data lookup
        model_type = "art"

        # Determine output directory (parent of model file for single-file models, or model_path for repos)
        single_file_extensions = (
            ".gguf",
            ".safetensors",
            ".ckpt",
            ".pt",
            ".pth",
        )
        if model_path and model_path.lower().endswith(single_file_extensions):
            output_dir = os.path.dirname(model_path)
        else:
            output_dir = model_path

        # Show download dialog (modal to ensure it stays visible)
        main_window = self._get_main_window()
        if main_window:
            try:
                # Close any existing download dialog
                if self._download_dialog:
                    self._download_dialog.close()
                    self._download_dialog = None

                # Create new download dialog and store reference
                self._download_dialog = HuggingFaceDownloadDialog(
                    parent=main_window,
                    model_name=f"{version} Config Files",
                    model_path=output_dir,
                )

                # Connect dialog to download worker signals
                self.huggingface_download_worker.register(
                    SignalCode.UPDATE_DOWNLOAD_LOG,
                    self._download_dialog.on_log_updated,
                )
                self.huggingface_download_worker.register(
                    SignalCode.UPDATE_DOWNLOAD_PROGRESS,
                    self._download_dialog.on_progress_updated,
                )
                self.huggingface_download_worker.register(
                    SignalCode.UPDATE_FILE_DOWNLOAD_PROGRESS,
                    self._download_dialog.on_file_progress_updated,
                )
                self.huggingface_download_worker.register(
                    SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                    self._download_dialog.on_download_complete,
                )
                self.huggingface_download_worker.register(
                    SignalCode.HUGGINGFACE_DOWNLOAD_FAILED,
                    self._download_dialog.on_download_failed,
                )

                # Show the dialog
                self._download_dialog.show()
                self._download_dialog.raise_()
                self._download_dialog.activateWindow()

                if self.logger:
                    self.logger.info(
                        "Download dialog shown and connected to worker signals"
                    )
            except Exception as e:
                if self.logger:
                    self.logger.error(
                        f"Error showing download dialog: {e}", exc_info=True
                    )
        else:
            if self.logger:
                self.logger.error(
                    "Unable to locate the main window; download dialog could not be displayed"
                )

        # Queue download request
        download_data = {
            "repo_id": repo_id,
            "model_type": model_type,
            "version": version,  # Pass full version name for bootstrap data lookup
            "pipeline_action": pipeline_action,  # Pass pipeline action (txt2img, inpaint, etc.)
            "output_dir": output_dir,
            "missing_files": missing_files,  # Pass specific missing files to download
        }

        if self.logger:
            self.logger.info(f"Queueing download request: {download_data}")

        self.huggingface_download_worker.add_to_queue(download_data)

    def _get_main_window(self):
        """Return the main application window if available."""
        try:
            from PySide6.QtWidgets import QApplication
        except ImportError:
            return None

        app = QApplication.instance()
        if app is None:
            return None

        window = app.activeWindow()
        if window is not None:
            return window

        for widget in app.topLevelWidgets():
            if (
                widget.objectName() == "MainWindow"
                or widget.__class__.__name__ == "MainWindow"
            ):
                return widget

        # Fallback: some components expose the main window via settings API
        return getattr(self.api, "main_window", None)

    def on_huggingface_download_complete(self, data: Dict):
        """Handle download completion and retry pending generation.

        Args:
            data: Download completion info with model_path
        """
        # Skip if we're in a batch download (OpenVoice handles its own completion)
        if hasattr(self, "_openvoice_remaining_downloads") and self._openvoice_remaining_downloads is not None:
            if self.logger:
                self.logger.debug(
                    f"Skipping generic download complete handler during batch download"
                )
            return
        
        # Handle safety checker downloads - notify via queue to ensure thread safety
        pipeline_action = data.get("pipeline_action", "")
        if pipeline_action == "safety_checker":
            if self.logger:
                self.logger.info(
                    "Safety checker download complete, notifying worker to retry load"
                )
            # Use queue to ensure thread-safe notification
            if self._safety_checker_worker is not None:
                self._safety_checker_worker.add_to_queue({"action": "load", "data": {}})
            return
        
        if self.logger:
            self.logger.info(
                f"WorkerManager: Download complete for {data.get('model_path')}"
            )
            self.logger.info(
                f"Pending generation request: {self._pending_generation_request is not None}"
            )

        # Delegate to LLM worker only for LLM model downloads
        # Skip TTS, STT, and other non-LLM model types
        # Use add_to_queue to ensure processing happens in worker thread,
        # not main thread, to prevent UI lockups during model loading
        model_type = data.get("model_type", "")
        non_llm_types = {"tts_openvoice", "stt", "openvoice_zip", "art"}
        if self._llm_generate_worker is not None and model_type not in non_llm_types:
            self.llm_generate_worker.add_to_queue({
                "_message_type": "download_complete",
                "data": data
            })

        # If we have a pending generation request (for image generation), retry it now
        if self._pending_generation_request:
            if self.logger:
                self.logger.info(
                    "Retrying pending generation request after download completion"
                )
                self.logger.debug(
                    f"Image request details: {self._pending_generation_request}"
                )

            # Re-emit the generation signal to trigger model loading and generation
            # The pending request already has the correct structure
            self.emit_signal(
                SignalCode.DO_GENERATE_SIGNAL,
                self._pending_generation_request,
            )
            self._pending_generation_request = None
        else:
            if self.logger:
                self.logger.warning(
                    "No pending generation request found after download completion"
                )

    def on_unload_art_signal(self, data: Dict):
        if self._sd_worker is not None:

            def callback(res: Dict):
                del self._sd_worker
                self._sd_worker = None

            data["callback"] = callback
            self.sd_worker.unload(data)
            self.sd_worker.image_export_worker.stop()
            del self.sd_worker.image_export_worker
            self.sd_worker.image_export_worker = None
            self._image_export_worker = None

    def on_sd_cancel_signal(self, data):
        if self._sd_worker is not None:
            self.sd_worker.on_sd_cancel_signal(data)

    def on_stop_auto_image_generation_signal(self, data):
        if self._sd_worker is not None:
            self.sd_worker.on_stop_auto_image_generation_signal(data)

    def on_interrupt_image_generation_signal(self, data):
        if self._sd_worker is not None:
            self.sd_worker.on_interrupt_image_generation_signal(data)

    def on_change_scheduler_signal(self, data):
        if self._sd_worker is not None:
            self.sd_worker.on_change_scheduler_signal(data)

    def on_model_status_changed_signal(self, data):
        from airunner.enums import ModelStatus, ModelType

        model_type = data.get("model") or data.get("model_type")
        status = data.get("status")

        if (
            model_type == ModelType.SAFETY_CHECKER
            and status == ModelStatus.LOADED
            and self._pending_generation_request is not None
        ):
            self.logger.info(
                "Safety checker loaded, proceeding with pending image generation"
            )
            # Note: Don't clear _pending_generation_request here - it will be cleared
            # by on_huggingface_download_complete after the SD model download completes,
            # or it will be set to None if no download is needed (generation proceeds immediately)
            self._proceed_with_generation(self._pending_generation_request)

        elif (
            model_type == ModelType.SAFETY_CHECKER
            and status == ModelStatus.FAILED
            and self._pending_generation_request is not None
        ):
            # Check if this was a download-triggered failure (we should wait for download)
            # or a real failure (show error)
            # The safety checker worker emits FAILED only after download fails or load fails
            # If NSFW filter is enabled, we should NOT proceed - show error instead
            from airunner.components.settings.data.application_settings import (
                ApplicationSettings,
            )
            app_settings = ApplicationSettings.objects.first()
            
            if app_settings.nsfw_filter:
                self.logger.error(
                    "Safety checker failed to load and NSFW filter is enabled. Cannot proceed with generation."
                )
                # Clear pending request since we're not proceeding
                self._pending_generation_request = None
                # Show error popup to user
                self.api.application_error(
                    message="Safety checker failed to load. Please disable the NSFW filter in settings or wait for the safety checker to download."
                )
            else:
                self.logger.warning(
                    "Safety checker failed to load; continuing image generation (NSFW filter disabled)"
                )
                self._proceed_with_generation(self._pending_generation_request)

        if self._sd_worker is not None:
            self.sd_worker.on_model_status_changed_signal(data)
        if self._stt_audio_capture_worker is not None:
            self.stt_audio_capture_worker.on_model_status_changed_signal(data)

    def on_load_art_signal(self, data):
        if self.sd_worker is not None:
            self.sd_worker.on_load_art_signal(data)

    def on_art_model_changed(self, data):
        if self._sd_worker is not None:
            self.sd_worker.on_art_model_changed(data)

    def on_load_controlnet_signal(self, data):
        if self._sd_worker is not None:
            self.sd_worker.on_load_controlnet_signal(data)

    def on_unload_controlnet_signal(self, data):
        if self._sd_worker is not None:
            self.sd_worker.on_unload_controlnet_signal(data)

    def on_safety_checker_load_signal(self, data):
        # Ensure the worker is created and send load request through its queue
        # Using add_to_queue ensures thread-safe message passing
        self.safety_checker_worker.add_to_queue({"action": "load", "data": data})

    def on_safety_checker_unload_signal(self, data):
        # Trigger unloading if worker exists
        if self._safety_checker_worker is not None:
            self._safety_checker_worker.handle_unload(data)

    def _handle_image_generation_request(self, data):
        """
        Handle image generation request, ensuring safety checker is ready if needed.

        Args:
            data: Image generation request data
        """
        from airunner.components.settings.data.application_settings import (
            ApplicationSettings,
        )

        app_settings = ApplicationSettings.objects.first()

        # Check if safety checker is enabled
        if not app_settings.nsfw_filter:
            # Safety checker disabled, proceed immediately
            self._proceed_with_generation(data)
            return

        # Safety checker is enabled, check if it's already loaded
        safety_worker = self.safety_checker_worker
        if (
            safety_worker
            and safety_worker.safety_checker is not None
            and safety_worker.feature_extractor is not None
        ):
            # Already loaded, proceed immediately
            self._proceed_with_generation(data)
            return

        # Safety checker needs to be loaded, store the request and trigger load
        self.logger.info(
            "Safety checker not ready, triggering load before generation"
        )
        self._pending_generation_request = data
        self.emit_signal(SignalCode.SAFETY_CHECKER_LOAD_SIGNAL, {})

    def _proceed_with_generation(self, data):
        """
        Proceed with image generation after safety checker is ready (or not needed).

        Args:
            data: Image generation request data
        """
        self.logger.info(f"_proceed_with_generation called with data keys: {data.keys() if data else 'None'}")
        if self.sd_worker is not None:
            self.logger.info("Calling sd_worker.on_do_generate_signal")
            self.sd_worker.on_do_generate_signal(data)
        else:
            self.logger.error("sd_worker is None, cannot proceed with generation")

    def on_input_image_settings_changed_signal(self, data):
        if self._sd_worker is not None:
            self.sd_worker.on_input_image_settings_changed_signal(data)

    def on_update_lora_signal(self, data):
        if self._sd_worker is not None:
            self.sd_worker.on_update_lora_signal(data)

    def on_update_embeddings_signal(self, data):
        if self._sd_worker is not None:
            self.sd_worker.on_update_embeddings_signal(data)

    def delete_missing_embeddings(self, data):
        if self._sd_worker is not None:
            self.sd_worker.delete_missing_embeddings(data)

    def on_llm_on_unload_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_llm_on_unload_signal(data)

    def on_llm_load_model_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_llm_load_model_signal(data)

    def on_fara_load_signal(self, data: Dict):
        """Handle FARA load signal.

        Unloads all other models (LLM, SD, TTS, STT) before loading Fara
        to ensure sufficient VRAM is available.

        Args:
            data: Signal data dictionary
        """
        if self.logger:
            self.logger.info("FARA_LOAD_SIGNAL received - unloading all models first")

        # Unload regular LLM if loaded
        if self._llm_generate_worker is not None:
            try:
                self.llm_generate_worker.unload_llm()
                if self.logger:
                    self.logger.info("Unloaded regular LLM model")
            except Exception as e:
                if self.logger:
                    self.logger.warning(f"Error unloading LLM: {e}")

        # Unload SD/art model if loaded
        if self._sd_worker is not None:
            try:
                self.sd_worker.on_unload_art_signal()
                if self.logger:
                    self.logger.info("Unloaded art/SD model")
            except Exception as e:
                if self.logger:
                    self.logger.warning(f"Error unloading SD: {e}")

        # Unload TTS if loaded
        if self.tts_generator_worker is not None:
            try:
                self.tts_generator_worker.unload()
                if self.logger:
                    self.logger.info("Unloaded TTS model")
            except Exception as e:
                if self.logger:
                    self.logger.warning(f"Error unloading TTS: {e}")

        # Unload STT if loaded
        if self._stt_audio_processor_worker is not None:
            try:
                self.stt_audio_processor_worker.unload()
                if self.logger:
                    self.logger.info("Unloaded STT model")
            except Exception as e:
                if self.logger:
                    self.logger.warning(f"Error unloading STT: {e}")

        # Now load Fara
        if self.fara_worker is not None:
            self.fara_worker.load(data)

    def on_fara_unload_signal(self, data: Dict):
        """Handle FARA unload signal.

        Args:
            data: Signal data dictionary
        """
        if self.logger:
            self.logger.info("FARA_UNLOAD_SIGNAL received")

        if self._fara_worker is not None:
            self.fara_worker.unload(data)

    def on_fara_model_download_required_signal(self, data: Dict):
        """Handle FARA model download required signal.

        Args:
            data: Signal data dictionary with model_path, model_name, repo_id
        """
        if self.logger:
            self.logger.info(f"FARA_MODEL_DOWNLOAD_REQUIRED received: {data}")

        if self._fara_worker is not None:
            self.fara_worker.on_fara_model_download_required_signal(data)

    def on_llm_model_changed_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_llm_model_changed_signal(data)

    def on_llm_model_download_required_signal(self, data):
        self.logger.info(f"WorkerManager received LLM_MODEL_DOWNLOAD_REQUIRED: {data}")
        if self._llm_generate_worker is not None:
            self.logger.info("Forwarding to llm_generate_worker")
            self.llm_generate_worker.on_llm_model_download_required_signal(
                data
            )
        else:
            # Handle download directly without LLM worker
            self.logger.info("LLM worker not available, handling download directly")
            self._handle_llm_download_directly(data)

    def _handle_llm_download_directly(self, data):
        """Handle LLM model download when LLM worker is not available.
        
        This allows downloading models from settings before the LLM is loaded.
        """
        from PySide6.QtWidgets import QApplication
        from airunner.components.llm.gui.windows.huggingface_download_dialog import (
            HuggingFaceDownloadDialog,
        )
        from airunner.components.llm.managers.download_huggingface import (
            DownloadHuggingFaceModel,
        )
        
        model_path = data.get("model_path", "")
        model_name = data.get("model_name", "Unknown Model")
        repo_id = data.get("repo_id", "")
        model_type = data.get("model_type", "llm")
        gguf_filename = data.get("gguf_filename")
        
        if not repo_id:
            self.logger.error("No repo_id provided in download request")
            return
        
        # Get main window
        main_window = None
        app = QApplication.instance()
        for widget in app.topLevelWidgets():
            if widget.__class__.__name__ == "MainWindow":
                main_window = widget
                break
        
        if not main_window:
            self.logger.error("Cannot show download dialog - main window not found")
            return
        
        is_gguf = model_type == "gguf" or gguf_filename is not None
        dialog_model_name = f"{model_name} (GGUF)" if is_gguf else model_name
        
        # Create and show download dialog
        self._download_dialog = HuggingFaceDownloadDialog(
            parent=main_window,
            model_name=dialog_model_name,
            model_path=model_path,
        )
        
        # Create download worker - use local variable to avoid polluting the 
        # huggingface_download_worker property which is used for STT/TTS downloads
        llm_download_worker = create_worker(DownloadHuggingFaceModel)
        
        # Connect dialog to download worker signals
        llm_download_worker.register(
            SignalCode.UPDATE_DOWNLOAD_LOG,
            self._download_dialog.on_log_updated,
        )
        llm_download_worker.register(
            SignalCode.UPDATE_DOWNLOAD_PROGRESS,
            self._download_dialog.on_progress_updated,
        )
        llm_download_worker.register(
            SignalCode.UPDATE_FILE_DOWNLOAD_PROGRESS,
            self._download_dialog.on_file_progress_updated,
        )
        llm_download_worker.register(
            SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
            self._download_dialog.on_download_complete,
        )
        llm_download_worker.register(
            SignalCode.HUGGINGFACE_DOWNLOAD_FAILED,
            self._download_dialog.on_download_failed,
        )
        
        if is_gguf and gguf_filename:
            self.logger.info(f"Starting GGUF download: {repo_id}/{gguf_filename}")
            llm_download_worker.download(
                repo_id=repo_id,
                model_type="gguf",
                output_dir=model_path,
                setup_quantization=False,
                quantization_bits=0,
                missing_files=None,
                gguf_filename=gguf_filename,
            )
        else:
            self.logger.info(f"Starting standard download: {repo_id}")
            llm_download_worker.download(
                repo_id=repo_id,
                model_type=model_type,
                output_dir=model_path,
                setup_quantization=True,
                quantization_bits=data.get("quantization_bits", 4),
                missing_files=data.get("missing_files"),
            )
        
        self._download_dialog.show()

    def on_huggingface_download_complete_signal(self, data):
        # Use add_to_queue to ensure processing happens in worker thread,
        # not main thread, to prevent UI lockups during model loading
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.add_to_queue({
                "_message_type": "download_complete",
                "data": data
            })

    def on_llm_reload_rag_index_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_llm_reload_rag_index_signal(data)

    def on_rag_index_all_documents_signal(self, data):
        if self.llm_generate_worker is not None:
            self.llm_generate_worker.on_rag_index_all_documents_signal(data)

    def on_rag_index_selected_documents_signal(self, data):
        if self.llm_generate_worker is not None:
            self.llm_generate_worker.on_rag_index_selected_documents_signal(
                data
            )

    def on_rag_index_cancel_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_rag_index_cancel_signal(data)

    def on_rag_load_documents_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_rag_load_documents_signal(data)

    def on_index_document_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_index_document_signal(data)

    def on_llm_start_fine_tune_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_llm_start_fine_tune_signal(data)

    def on_llm_fine_tune_cancel_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_llm_fine_tune_cancel_signal(data)

    def on_llm_start_quantization_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_llm_start_quantization_signal(data)

    def on_llm_clear_history_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_llm_clear_history_signal(data)

    def on_llm_add_chatbot_response_to_history(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_llm_add_chatbot_response_to_history(
                data
            )

    def on_llm_load_conversation(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_llm_load_conversation(data)

    def on_quit_application_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_quit_application_signal(data)

    def on_conversation_deleted_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_conversation_deleted_signal(data)

    def on_section_changed_signal(self, data):
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.on_section_changed_signal(data)

    def on_generate_mask_signal(self, data):
        if self._mask_generator_worker is not None:
            self.mask_generator_worker.on_generate_mask_signal(data)

    def on_audio_capture_worker_response_signal(self, data):
        if self._stt_audio_capture_worker is not None:
            self.stt_audio_capture_worker.on_audio_capture_worker_response_signal(
                data
            )

    def on_stt_stop_capture_signal(self, data):
        if self._stt_audio_capture_worker is not None:
            self.stt_audio_capture_worker.on_stt_stop_capture_signal(data)

    def on_recording_device_changed_signal(self, data):
        if self._stt_audio_capture_worker is not None:
            self.stt_audio_capture_worker.on_recording_device_changed_signal(
                data
            )

    def update_properties(self, data):
        if self._stt_audio_processor_worker is not None:
            self.stt_audio_processor_worker.update_properties(data)

    def on_stt_unload_signal(self, data):
        if self._stt_audio_processor_worker is not None:
            self.stt_audio_processor_worker.on_stt_unload_signal(data)

    def on_stt_process_audio_signal(self, data):
        # Use the property to ensure the worker is created
        if self.logger:
            self.logger.debug("on_stt_process_audio_signal called")
        worker = self.stt_audio_processor_worker
        if worker is not None:
            if self.logger:
                self.logger.debug("Forwarding audio to stt_audio_processor_worker")
            worker.on_stt_process_audio_signal(data)
        else:
            if self.logger:
                self.logger.warning("stt_audio_processor_worker is None")

    def on_interrupt_process_signal(self, data):
        # Interrupt LLM generation
        if self._llm_generate_worker is not None:
            self.llm_generate_worker.llm_on_interrupt_process_signal(data)

        # Interrupt TTS generation
        if self.tts_generator_worker is not None:
            self.tts_generator_worker.on_interrupt_process_signal(data)

        if self._tts_vocalizer_worker is not None:
            self.tts_vocalizer_worker.on_interrupt_process_signal(data)

    def on_unblock_tts_generator_signal(self, data):
        callback_handled = False
        
        if self.tts_generator_worker is not None:
            self.tts_generator_worker.on_unblock_tts_generator_signal(data)
            callback_handled = True

        if self._tts_vocalizer_worker is not None:
            self.tts_vocalizer_worker.on_unblock_tts_generator_signal(data)
            callback_handled = True

        # If no TTS workers handled this, still invoke the callback
        # This ensures STT audio capture works even when TTS is disabled
        if not callback_handled and data is not None:
            callback = data.get("callback", None)
            if callback is not None:
                callback()

    def on_disable_tts_signal(self, data):
        if self.tts_generator_worker is not None:
            self.tts_generator_worker.on_disable_tts_signal(data)

    def on_llm_text_streamed_signal(self, data):
        if self.tts_generator_worker is not None:
            self.tts_generator_worker.on_llm_text_streamed_signal(data)

    def _reload_tts_model_manager(self, data):
        if self.tts_generator_worker is not None:
            self.tts_generator_worker._reload_tts_model_manager(data)

    def on_application_settings_changed_signal(self, data):
        if self.tts_generator_worker is not None:
            self.tts_generator_worker.on_application_settings_changed_signal(
                data
            )

        if self._tts_vocalizer_worker is not None:
            self.tts_vocalizer_worker.on_application_settings_changed_signal(
                data
            )

    def on_add_to_queue_signal(self, data):
        if self.tts_generator_worker is not None:
            self.tts_generator_worker.on_add_to_queue_signal(data)

    def on_playback_device_changed_signal(self, data):
        if self._tts_vocalizer_worker is not None:
            self.tts_vocalizer_worker.on_playback_device_changed_signal(data)

    def on_image_exported_signal(self, data):
        # self.api.art.unload()
        pass

    def on_start_huggingface_download_signal(self, data: dict):
        """Handle START_HUGGINGFACE_DOWNLOAD signal for STT/TTS model downloads.

        This signal is emitted when model managers detect missing model files
        and need to trigger a download. Uses the same download infrastructure
        as LLM and Art models.

        Args:
            data: Signal data dictionary containing:
                - repo_id: HuggingFace repository ID
                - model_path: Local path to save the model
                - model_type: Type of model (stt, tts_openvoice)
                - callback: Optional callback to invoke after download completes
        """
        from PySide6.QtWidgets import QApplication, QMessageBox
        from airunner.components.llm.gui.windows.huggingface_download_dialog import (
            HuggingFaceDownloadDialog,
        )
        
        # Check if HuggingFace downloads are allowed
        from airunner.components.application.gui.dialogs.privacy_consent_dialog import (
            is_huggingface_allowed,
        )
        if not is_huggingface_allowed():
            if self.logger:
                self.logger.info("HuggingFace downloads disabled by privacy settings")
            main_window = self._get_main_window()
            if main_window:
                QMessageBox.warning(
                    main_window,
                    "Downloads Disabled",
                    "HuggingFace downloads are disabled in privacy settings.\n\n"
                    "You can enable them in Preferences → Privacy & Security → External Services."
                )
            return

        repo_id = data.get("repo_id", "")
        model_path = data.get("model_path", "")
        model_type = data.get("model_type", "")
        callback = data.get("callback")

        if not repo_id:
            if self.logger:
                self.logger.error("No repo_id provided in download request")
            return

        if self.logger:
            self.logger.info(
                f"Starting HuggingFace download for {model_type}: {repo_id}"
            )

        # Get main window for dialog parent
        main_window = self._get_main_window()

        if main_window:
            try:
                # Close any existing download dialog
                if hasattr(self, "_download_dialog") and self._download_dialog:
                    self._download_dialog.close()
                    self._download_dialog = None

                # Create download dialog
                model_name = repo_id.split("/")[-1] if "/" in repo_id else repo_id
                self._download_dialog = HuggingFaceDownloadDialog(
                    parent=main_window,
                    model_name=f"{model_name} ({model_type.upper()})",
                    model_path=model_path,
                )

                # Store the callback for after download completes
                self._pending_download_callback = callback

                # Connect dialog to download worker signals
                self.huggingface_download_worker.register(
                    SignalCode.UPDATE_DOWNLOAD_LOG,
                    self._download_dialog.on_log_updated,
                )
                self.huggingface_download_worker.register(
                    SignalCode.UPDATE_DOWNLOAD_PROGRESS,
                    self._download_dialog.on_progress_updated,
                )
                self.huggingface_download_worker.register(
                    SignalCode.UPDATE_FILE_DOWNLOAD_PROGRESS,
                    self._download_dialog.on_file_progress_updated,
                )
                self.huggingface_download_worker.register(
                    SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                    self._on_stt_tts_download_complete,
                )
                self.huggingface_download_worker.register(
                    SignalCode.HUGGINGFACE_DOWNLOAD_FAILED,
                    self._download_dialog.on_download_failed,
                )

                # Show the dialog
                self._download_dialog.show()
                self._download_dialog.raise_()
                self._download_dialog.activateWindow()

                if self.logger:
                    self.logger.info(
                        "Download dialog shown for STT/TTS model"
                    )

            except Exception as e:
                if self.logger:
                    self.logger.error(
                        f"Error showing download dialog: {e}", exc_info=True
                    )
        else:
            if self.logger:
                self.logger.error(
                    "Unable to locate the main window; download dialog could not be displayed"
                )

        # Queue download request
        download_data = {
            "repo_id": repo_id,
            "model_type": model_type,
            "output_dir": model_path,
        }

        if self.logger:
            self.logger.info(f"Queueing download request: {download_data}")

        self.huggingface_download_worker.add_to_queue(download_data)

    def _on_stt_tts_download_complete(self, data: dict):
        """Handle completion of STT/TTS model download.

        Invokes the stored callback if one was provided.
        """
        if hasattr(self, "_download_dialog") and self._download_dialog:
            self._download_dialog.on_download_complete(data)

        # Invoke callback if provided
        if hasattr(self, "_pending_download_callback") and self._pending_download_callback:
            callback = self._pending_download_callback
            self._pending_download_callback = None
            try:
                callback()
            except Exception as e:
                if self.logger:
                    self.logger.error(
                        f"Error invoking download callback: {e}", exc_info=True
                    )

    def on_start_openvoice_zip_download_signal(self, data: dict):
        """Handle START_OPENVOICE_ZIP_DOWNLOAD signal for OpenVoice converter checkpoints.

        Uses the same download dialog as HuggingFace downloads.

        Args:
            data: Signal data dictionary containing:
                - callback: Optional callback to invoke after download completes
        """
        import os
        from PySide6.QtWidgets import QApplication
        from airunner.components.llm.gui.windows.huggingface_download_dialog import (
            HuggingFaceDownloadDialog,
        )

        callback = data.get("callback")
        
        # Target directory for extraction
        openvoice_dir = os.path.join(
            self.path_settings.tts_model_path,
            "openvoice",
        )
        
        if self.logger:
            self.logger.info(
                f"Starting OpenVoice ZIP download to {openvoice_dir}"
            )

        # Get main window for dialog parent
        main_window = self._get_main_window()

        if main_window:
            try:
                # Close any existing download dialog
                if hasattr(self, "_download_dialog") and self._download_dialog:
                    self._download_dialog.close()
                    self._download_dialog = None

                # Create download dialog
                self._download_dialog = HuggingFaceDownloadDialog(
                    parent=main_window,
                    model_name="OpenVoice Converter Checkpoints",
                    model_path=openvoice_dir,
                )

                # Store the callback for after download completes
                self._pending_download_callback = callback

                # Connect dialog to download worker signals
                self.huggingface_download_worker.register(
                    SignalCode.UPDATE_DOWNLOAD_LOG,
                    self._download_dialog.on_log_updated,
                )
                self.huggingface_download_worker.register(
                    SignalCode.UPDATE_DOWNLOAD_PROGRESS,
                    self._download_dialog.on_progress_updated,
                )
                self.huggingface_download_worker.register(
                    SignalCode.UPDATE_FILE_DOWNLOAD_PROGRESS,
                    self._download_dialog.on_file_progress_updated,
                )
                self.huggingface_download_worker.register(
                    SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                    self._on_stt_tts_download_complete,
                )
                self.huggingface_download_worker.register(
                    SignalCode.HUGGINGFACE_DOWNLOAD_FAILED,
                    self._download_dialog.on_download_failed,
                )

                # Show the dialog
                self._download_dialog.show()
                self._download_dialog.raise_()
                self._download_dialog.activateWindow()

                if self.logger:
                    self.logger.info(
                        "Download dialog shown for OpenVoice checkpoints"
                    )

            except Exception as e:
                if self.logger:
                    self.logger.error(
                        f"Error showing download dialog: {e}", exc_info=True
                    )
        else:
            if self.logger:
                self.logger.error(
                    "Unable to locate the main window; download dialog could not be displayed"
                )

        # Queue download request - use the openvoice_zip model type
        download_data = {
            "model_type": "openvoice_zip",
            "output_dir": openvoice_dir,
            "zip_url": "https://myshell-public-repo-host.s3.amazonaws.com/openvoice/checkpoints_v2_0417.zip",
        }

        if self.logger:
            self.logger.info(f"Queueing OpenVoice ZIP download request: {download_data}")

        self.huggingface_download_worker.add_to_queue(download_data)

    def on_start_openvoice_batch_download_signal(self, data: dict):
        """Handle START_OPENVOICE_BATCH_DOWNLOAD signal.

        Shows a language selection dialog FIRST, then downloads all selected models
        (including converter ZIP and HuggingFace models) in a single batch.

        Args:
            data: Signal data containing:
                - needs_converter: Whether the converter ZIP needs download
                - missing_core_models: List of core model IDs that need download
                - missing_languages: List of language keys that need download
                - callback: Callback to invoke after all downloads complete
        """
        import os
        from PySide6.QtWidgets import QMessageBox, QDialog
        from airunner.components.tts.gui.dialogs.openvoice_language_dialog import (
            OpenVoiceLanguageDialog,
        )
        from airunner.components.tts.data.bootstrap.openvoice_languages import (
            OPENVOICE_CORE_MODELS,
            OPENVOICE_LANGUAGE_MODELS,
            get_models_for_languages,
        )
        from airunner.components.llm.gui.windows.huggingface_download_dialog import (
            HuggingFaceDownloadDialog,
        )

        needs_converter = data.get("needs_converter", False)
        missing_core_models = data.get("missing_core_models", [])
        missing_languages = data.get("missing_languages", [])
        callback = data.get("callback")

        if self.logger:
            self.logger.info(
                f"OpenVoice batch download requested: converter={needs_converter}, "
                f"{len(missing_core_models)} core models, {len(missing_languages)} languages"
            )

        # Clean up any previous batch download state and ALL other download handlers
        # that might interfere with the batch download
        try:
            self.huggingface_download_worker.unregister(
                SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                self._on_openvoice_batch_download_complete,
            )
        except Exception:
            pass
        # Unregister the STT/TTS download complete handler if it exists
        try:
            self.huggingface_download_worker.unregister(
                SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                self._on_stt_tts_download_complete,
            )
        except Exception:
            pass
        # Unregister any direct dialog handlers from other download flows
        if hasattr(self, "_download_dialog") and self._download_dialog:
            try:
                self.huggingface_download_worker.unregister(
                    SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                    self._download_dialog.on_download_complete,
                )
            except Exception:
                pass
        self._openvoice_remaining_downloads = None
        self._openvoice_download_callback = None
        self._pending_download_callback = None  # Clear any pending STT/TTS callback

        main_window = self._get_main_window()
        if not main_window:
            if self.logger:
                self.logger.error("Cannot show language dialog - no main window")
            return

        # Show language selection dialog FIRST
        # This lets the user choose languages before any downloads start
        dialog = OpenVoiceLanguageDialog(
            parent=main_window,
            missing_languages=missing_languages,
        )
        result = dialog.exec()
        
        if result != QDialog.Accepted:
            if self.logger:
                self.logger.info("User cancelled OpenVoice language selection")
            return
        
        selected_languages = dialog.get_selected_languages()

        # Build list of all models to download
        models_to_download = list(missing_core_models)  # Always download core models
        
        # Add language-specific models for selected languages
        for lang_key in selected_languages:
            if lang_key in OPENVOICE_LANGUAGE_MODELS:
                lang_info = OPENVOICE_LANGUAGE_MODELS[lang_key]
                for model_id in lang_info["models"]:
                    if model_id not in models_to_download:
                        models_to_download.append(model_id)

        # Count total downloads (including converter if needed)
        total_downloads = len(models_to_download)
        if needs_converter:
            total_downloads += 1

        if total_downloads == 0:
            if self.logger:
                self.logger.info("No OpenVoice models to download")
            # Still invoke callback since download check passed
            if callback:
                callback()
            return

        if self.logger:
            self.logger.info(
                f"Downloading {total_downloads} OpenVoice items: "
                f"converter={needs_converter}, models={models_to_download}"
            )

        # Create download dialog
        try:
            # Clean up any existing dialog and its signal handlers
            if hasattr(self, "_download_dialog") and self._download_dialog:
                # Unregister all signals connected to the old dialog
                try:
                    self.huggingface_download_worker.unregister(
                        SignalCode.UPDATE_DOWNLOAD_LOG,
                        self._download_dialog.on_log_updated,
                    )
                except Exception:
                    pass
                try:
                    self.huggingface_download_worker.unregister(
                        SignalCode.UPDATE_DOWNLOAD_PROGRESS,
                        self._download_dialog.on_progress_updated,
                    )
                except Exception:
                    pass
                try:
                    self.huggingface_download_worker.unregister(
                        SignalCode.UPDATE_FILE_DOWNLOAD_PROGRESS,
                        self._download_dialog.on_file_progress_updated,
                    )
                except Exception:
                    pass
                try:
                    self.huggingface_download_worker.unregister(
                        SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                        self._download_dialog.on_download_complete,
                    )
                except Exception:
                    pass
                try:
                    self.huggingface_download_worker.unregister(
                        SignalCode.HUGGINGFACE_DOWNLOAD_FAILED,
                        self._download_dialog.on_download_failed,
                    )
                except Exception:
                    pass
                self._download_dialog.close()
                self._download_dialog = None

            self._download_dialog = HuggingFaceDownloadDialog(
                parent=main_window,
                model_name=f"OpenVoice TTS ({total_downloads} items)",
                model_path=os.path.join(
                    self.path_settings.base_path, "text/models/tts"
                ),
                batch_mode=True,  # Batch mode: don't auto-close on each download
            )

            # Track remaining downloads and callback
            self._openvoice_remaining_downloads = total_downloads
            self._openvoice_download_callback = callback

            # Connect signals
            self.huggingface_download_worker.register(
                SignalCode.UPDATE_DOWNLOAD_LOG,
                self._download_dialog.on_log_updated,
            )
            self.huggingface_download_worker.register(
                SignalCode.UPDATE_DOWNLOAD_PROGRESS,
                self._download_dialog.on_progress_updated,
            )
            self.huggingface_download_worker.register(
                SignalCode.UPDATE_FILE_DOWNLOAD_PROGRESS,
                self._download_dialog.on_file_progress_updated,
            )
            self.huggingface_download_worker.register(
                SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                self._on_openvoice_batch_download_complete,
            )
            self.huggingface_download_worker.register(
                SignalCode.HUGGINGFACE_DOWNLOAD_FAILED,
                self._download_dialog.on_download_failed,
            )

            # Show dialog
            self._download_dialog.show()
            self._download_dialog.raise_()
            self._download_dialog.activateWindow()

            # Queue converter ZIP download first if needed
            if needs_converter:
                openvoice_dir = os.path.join(
                    self.path_settings.base_path, "text/models/tts/openvoice"
                )
                os.makedirs(openvoice_dir, exist_ok=True)
                
                download_data = {
                    "model_type": "openvoice_zip",
                    "output_dir": openvoice_dir,
                    "zip_url": "https://myshell-public-repo-host.s3.amazonaws.com/openvoice/checkpoints_v2_0417.zip",
                }
                if self.logger:
                    self.logger.info(f"Queueing OpenVoice converter ZIP download")
                self.huggingface_download_worker.add_to_queue(download_data)

            # Queue all HuggingFace model downloads
            for model_id in models_to_download:
                model_path = os.path.join(
                    self.path_settings.base_path,
                    "text/models/tts",
                    model_id,
                )
                download_data = {
                    "repo_id": model_id,
                    "model_type": "tts_openvoice",
                    "output_dir": model_path,
                }
                self.huggingface_download_worker.add_to_queue(download_data)

        except Exception as e:
            if self.logger:
                self.logger.error(
                    f"Error setting up OpenVoice batch download: {e}", exc_info=True
                )

    def _on_openvoice_batch_download_complete(self, data: dict):
        """Handle completion of a single model in the OpenVoice batch download."""
        if not hasattr(self, "_openvoice_remaining_downloads") or self._openvoice_remaining_downloads is None:
            # Not in a batch download, ignore
            return
            
        self._openvoice_remaining_downloads -= 1
        
        if self.logger:
            self.logger.info(
                f"OpenVoice model download complete, {self._openvoice_remaining_downloads} remaining"
            )
        
        if self._openvoice_remaining_downloads <= 0:
            # All downloads complete - unregister all handlers
            try:
                self.huggingface_download_worker.unregister(
                    SignalCode.HUGGINGFACE_DOWNLOAD_COMPLETE,
                    self._on_openvoice_batch_download_complete,
                )
            except Exception:
                pass
            
            # Unregister dialog signal handlers
            if hasattr(self, "_download_dialog") and self._download_dialog:
                try:
                    self.huggingface_download_worker.unregister(
                        SignalCode.UPDATE_DOWNLOAD_LOG,
                        self._download_dialog.on_log_updated,
                    )
                except Exception:
                    pass
                try:
                    self.huggingface_download_worker.unregister(
                        SignalCode.UPDATE_DOWNLOAD_PROGRESS,
                        self._download_dialog.on_progress_updated,
                    )
                except Exception:
                    pass
                try:
                    self.huggingface_download_worker.unregister(
                        SignalCode.UPDATE_FILE_DOWNLOAD_PROGRESS,
                        self._download_dialog.on_file_progress_updated,
                    )
                except Exception:
                    pass
                try:
                    self.huggingface_download_worker.unregister(
                        SignalCode.HUGGINGFACE_DOWNLOAD_FAILED,
                        self._download_dialog.on_download_failed,
                    )
                except Exception:
                    pass
                
                # Notify dialog that all downloads are complete, then close it
                self._download_dialog.on_download_complete(data)
                # Use QTimer to close after a brief delay so user sees the completion message
                from PySide6.QtCore import QTimer
                dialog = self._download_dialog
                QTimer.singleShot(1500, dialog.accept)
                self._download_dialog = None
            
            # Clear tracking state
            self._openvoice_remaining_downloads = None
            
            # Invoke callback
            if hasattr(self, "_openvoice_download_callback") and self._openvoice_download_callback:
                callback = self._openvoice_download_callback
                self._openvoice_download_callback = None
                try:
                    callback()
                except Exception as e:
                    if self.logger:
                        self.logger.error(
                            f"Error invoking OpenVoice download callback: {e}",
                            exc_info=True
                        )



import os
import re
from sqlalchemy import Column, Integer, String, event
from sqlalchemy.orm import validates

from airunner.components.data.models.base import BaseModel
from airunner.settings import AIRUNNER_BASE_PATH


def _resolve_path(stored_path: str) -> str:
    """
    Resolve a stored path to work across different environments.
    
    Handles paths that were stored with a different home directory
    (e.g., /home/joe/.local/share/airunner stored on host, but running
    in Docker where home is /root).
    
    This allows the same database to work both on the host and in containers.
    """
    if not stored_path:
        return stored_path
    
    # Pattern to match paths like /home/<user>/.local/share/airunner/...
    # or /root/.local/share/airunner/...
    pattern = r'^(/home/[^/]+|/root)/\.local/share/airunner(/.*)?$'
    match = re.match(pattern, stored_path)
    
    if match:
        # Extract the relative part after .local/share/airunner
        relative_part = match.group(2) or ''
        # Reconstruct with current AIRUNNER_BASE_PATH
        return AIRUNNER_BASE_PATH + relative_part
    
    return stored_path


class PathSettings(BaseModel):
    """
    Path settings with automatic cross-environment path resolution.
    
    When paths are stored from one environment (e.g., /home/joe on host)
    and accessed from another (e.g., /root in Docker), they are automatically
    translated to work in the current environment.
    """
    __tablename__ = "path_settings"
    id = Column(Integer, primary_key=True, autoincrement=True)
    base_path = Column(String, default=AIRUNNER_BASE_PATH)
    documents_path = Column(
        String,
        default=os.path.expanduser(
            os.path.join(AIRUNNER_BASE_PATH, "text/other", "documents")
        ),
    )
    ebook_path = Column(
        String,
        default=os.path.expanduser(
            os.path.join(AIRUNNER_BASE_PATH, "text/other", "ebooks")
        ),
    )
    image_path = Column(
        String,
        default=os.path.expanduser(
            os.path.join(AIRUNNER_BASE_PATH, "art/other", "images")
        ),
    )
    llama_index_path = Column(
        String,
        default=os.path.expanduser(
            os.path.join(AIRUNNER_BASE_PATH, "text/rag", "db")
        ),
    )
    webpages_path = Column(
        String,
        default=os.path.expanduser(
            os.path.join(AIRUNNER_BASE_PATH, "text/other", "webpages")
        ),
    )
    stt_model_path = Column(
        String,
        default=os.path.expanduser(
            os.path.join(AIRUNNER_BASE_PATH, "text/models/stt")
        ),
    )
    tts_model_path = Column(
        String,
        default=os.path.expanduser(
            os.path.join(AIRUNNER_BASE_PATH, "text/models/tts")
        ),
    )

    # List of path attributes that should be auto-resolved
    _PATH_ATTRS = {
        'base_path', 'documents_path', 'ebook_path', 'image_path',
        'llama_index_path', 'webpages_path', 'stt_model_path', 'tts_model_path'
    }

    def __getattribute__(self, name: str):
        """Override attribute access to auto-resolve paths."""
        # Get the raw value first using parent's __getattribute__
        value = super().__getattribute__(name)
        
        # Check if this is a path attribute that needs resolution
        # Use object.__getattribute__ to avoid recursion
        path_attrs = object.__getattribute__(self, '_PATH_ATTRS')
        if name in path_attrs and isinstance(value, str):
            return _resolve_path(value)
        
        return value

    def tts_processor_path(self) -> str:
        return os.path.join(self.tts_model_path, "processor")



from setuptools import setup, find_packages

extras_require = {
    # These are optional dependencies that will change the
    # behavior of the application or add new features if installed.
    "nvidia": [  # NVIDIA dependencies:
        "nvidia-cuda-runtime",  # CUDA runtime (replaces deprecated cu12/cu13 variants)
    ],
    "huggingface": [
        "diffusers==0.35.1",
        "controlnet_aux==0.0.10",
        "safetensors==0.6.2",
        "compel==2.1.1",
        "transformers==4.57.1",
        "datasets==4.0.0",
        "peft==0.17.1",
    ],
    "gui": [  # GUI dependencies
        "PySide6==6.9.0",
        "PySide6_Addons==6.9.0",
        "PySide6_Essentials==6.9.0",
    ],
    "linux": [  # Linux-specific dependencies
        # "faiss-gpu==1.7.2", # If faiss-gpu is from NVIDIA or a custom index, it needs similar handling
        # "tensorrt==10.13.3.9",  # Temporarily disabled - depends on deprecated nvidia-cuda-runtime-cu13
    ],
    "dev": [  # Development dependencies
        "pytest",
        "pytest-timeout",
        "responses>=0.25.0",
        "python-dotenv==1.0.1",
        "coverage==7.8.0",
        "black==25.1.0",
        "pyinstaller==6.12.0",
        "flake8==7.2.0",
        "mypy==1.16.0",
        "autoflake==2.3.1",
        "pandas>=2.0.0",  # For eval dataset loading (parquet)
        "pyarrow>=14.0.0",  # For parquet file support
        "tqdm>=4.0.0",  # For progress bars in headless downloads
    ],
    "art": [  # Art generation dependencies
        "DeepCache==0.1.1",
        "tomesd==0.1.3",
        "gguf==0.17.1",
    ],
    "llm": [  # LLM dependencies (also text-to-speech and speech-to-text)
        "bitsandbytes==0.45.5",
        "sentence_transformers==3.4.1",
        "sounddevice==0.5.1",
        "pyttsx3==2.91",
        "cryptography==44.0.3",
        # Speech-to-text: faster-whisper (CTranslate2-based, ~4x faster than transformers)
        "faster-whisper>=1.0.0",
        # "flash_attn==2.7.4.post1", # flash-attn usually requires specific build steps.
        # GGUF model support via llama.cpp (smaller, faster than BitsAndBytes)
        # Note: For GPU support, install with: CMAKE_ARGS="-DGGML_CUDA=on" pip install llama-cpp-python
        "llama-cpp-python>=0.3.0",
        # Summarizations (basic)
        "sumy==0.11.0",
        "sentencepiece==0.2.0",
        "lingua-language-detector==2.1.0",
        "markdown==3.8",
        "libzim==3.7.0",
        # Mistral native function calling and Mistral3 tokenization
        "mistral_common>=1.8.5",
        # Tool search (BM25 ranking for on-demand tool discovery)
        "rank-bm25>=0.2.2",
        # llama-index (for RAG only)
        "llama-index-core>=0.13",
        "llama-index-readers-file>=0.5.0",
        "llama-index-embeddings-huggingface>=0.6.0",
        "llama-cloud>=0.1.0",
        # LangChain/LangGraph (for agent system)
        "langchain==1.0.0",
        "langchain-core==1.0.0",
        "langchain-community>=0.4.0",
        "langchain-huggingface>=0.1.0",
        "langgraph==1.0.0",
        "langsmith>=0.1.0",
        # Optional LangChain backends (commented out by default)
        # "langchain-openai>=0.2.0",  # For OpenRouter/OpenAI
        "langchain-ollama==1.0.0",  # For Ollama
        # "langchain-anthropic>=0.3.0",  # For Anthropic Claude
        # Document processing
        "EbookLib==0.19",
        "html2text==2025.4.15",
        "rake_nltk==1.0.6",
        "markdownify>=0.13.1",
    ],
    "llm_weather": [  # LLM dependencies for weather (requires llm dependencies)
        "requests-cache==1.2.1",
        "retry-requests==2.0.0",
        "openmeteo_requests==1.4.0",
        "uszipcode==1.0.1",
    ],
    "tts": [  # Text-to-speech dependencies (requires llm dependencies)
        "inflect==7.5.0",
        "pycountry==24.6.1",
        "librosa==0.11.0",
        "torchcodec>=0.8.0",  # Required by torchaudio for audio loading
    ],
    "rabbitmq": ["pika"],
    "openvoice": [
        "librosa==0.11.0",
        "pydub==0.25.1",
        "wavmark==0.0.3",
        "eng_to_ipa==0.0.2",
        "inflect==7.5.0",
        "unidecode==1.4.0",
        "langid==1.1.6",
    ],
    "melotts": [
        "txtsplit==1.0.0",
        "cached_path==1.7.3",
        "num2words==0.5.14",
        "g2p_en==2.1.0",
        "anyascii==0.3.2",
        "loguru==0.7.3",
    ],
    "openvoice_cn": [
        "pypinyin==0.54.0",
        "jieba==0.42.1",
        "cn2an==0.5.23",
    ],
    "openvoice_jp": [
        "unidic_lite==1.0.8",
        "unidic==1.1.0",
        "mecab-python3==1.0.10",
        # Note: fugashi requires MeCab to be installed on the system
        # Install with: apt-get install mecab libmecab-dev mecab-ipadic-utf8
        "fugashi==1.4.0",
        "pykakasi==2.3.0",
    ],
    "openvoice_kr": [
        "jamo==0.4.1",
        "python-mecab-ko==1.3.7",
        "python-mecab-ko-dic==2.1.1.post2",
    ],
    "openvoice_tw": [
        "g2pkk>=0.1.2",
    ],
    "gruut_support": [
        "gruut[de,es,fr]==2.4.0",
        "networkx==3.4.2",
    ],
    "search": [
        "ddgs>=9.0.0",  # Renamed from duckduckgo-search
        "aiohttp>=3.11.0",
        "google-api-python-client>=2.170.0",
        "wikipedia>=1.4.0",
        "scrapy==2.13.1",
        "trafilatura==2.0.0",
    ],
    "computer_use": [  # Computer use / desktop automation dependencies
        "pyautogui>=0.9.54",
        "pillow>=12.0.0",  # Already in install_requires, but listed for clarity
        "pyscreeze>=1.0.1",  # Screenshot backend for pyautogui
        "python-xlib>=0.33;platform_system=='Linux'",  # X11 support on Linux
        "pygetwindow>=0.0.9",  # Window management
    ],
}

extras_require["all"] = []
extras_require["all_dev"] = []
extras_require["windows"] = []

for k, v_list in extras_require.items():
    if k == "all":
        continue
    if k != "dev":
        extras_require["all"].extend(v_list)
    extras_require["all_dev"].extend(v_list)

for k, v in extras_require.items():
    if k in ["all", "all_dev", "windows"]:
        continue
    if k != "dev":
        if k not in extras_require["all"]:
            extras_require["all"].extend(v)
    if k not in extras_require["all_dev"]:
        extras_require["all_dev"].extend(v)
    if k != "linux":
        if k not in extras_require["windows"]:
            extras_require["windows"].extend(v)


setup(
    name="airunner",
    version="5.3.2",
    author="Capsize LLC",
    description="Run local opensource AI models (Stable Diffusion, LLMs, TTS, STT, chatbots) in a lightweight Python GUI",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="llm, pyside6, gui, local llm, stable diffusion, generative ai, local chatgpt, text-to-speech, speech-to-text, open source chatbot, python ai runner",
    license="Apache-2.0",
    author_email="contact@capsizegames.com",
    url="https://github.com/Capsize-Games/airunner",
    package_dir={"": "src"},
    packages=find_packages("src"),
    python_requires=">=3.13.3",
    install_requires=[
        "pip==25.3",
        "torch",
        "torchvision",
        "torchaudio",
        "torchao",
        "accelerate==1.7.0",
        "huggingface-hub>=0.24.0,<1.0",
        "tokenizers==0.22.0",
        "optimum==1.25.1",
        "numpy==2.2.5",
        "pillow==12.0.0",
        "alembic==1.15.2",
        "aiosqlite==0.21.0",
        "sqlalchemy==2.0.38",
        "setuptools==80.9.0",
        "etils[epath]==1.12.2",
        "jinja2==3.1.6",
        "pyyaml==6.0.2",
        "fastapi==0.115.0",
        "uvicorn[standard]==0.34.0",
    ],
    extras_require=extras_require,
    package_data={
        "airunner": [
            # Alembic migrations
            "alembic/*.py",
            "alembic/*.mako",
            "alembic/versions/*.py",
            "*.ini",
            # GUI resources
            "gui/cursors/*",
            "gui/images/*",
            "gui/resources/**/*",
            "gui/styles/**/*",
            # Component resources
            "components/icons/*",
            "components/art/filters/*",
            # UI templates (all .ui files in templates directories)
            "components/**/templates/*.ui",
            # Legal documents (user agreement, privacy policy)
            "components/**/user_agreement/*.md",
            # Static files (HTML, CSS, JS templates for web views)
            "components/**/static/**/*",
            "static/**/*",
            # Vendor resources
            "vendor/nodegraphqt/widgets/icons/*",
        ],
    },
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "airunner=airunner.launcher:main",
            "airunner-headless=airunner.bin.airunner_headless:main",
            "airunner-build-ui=airunner.bin.build_ui:main",
            "airunner-compile-translations=airunner.bin.compile_translations:main",
            "airunner-tests=airunner.bin.run_tests:main",
            "airunner-test-coverage-report=airunner.bin.coverage_report:main",
            "airunner-docker=airunner.bin.docker_wrapper:main",
            "airunner-generate-migration=airunner.bin.generate_migration:main",
            "airunner-generate-cert=airunner.bin.generate_cert:main",
            "airunner-mypy=airunner.bin.mypy_shortcut:main",
            "airunner-create-theme=airunner.bin.airunner_create_theme:main",
            "airunner-create-component=airunner.bin.airunner_create_component:main",
            "airunner-train-diffusers=airunner.bin.train_diffusers:main",
            "airunner-daemon=airunner.services.daemon:main",
            "airunner-service=airunner.bin.airunner_service:main",
            "airunner-quality-report=airunner.bin.code_quality_report:main",
            "airunner-remove-unused-imports=airunner.bin.remove_unused_imports:main",
            "airunner-migrate-knowledge=airunner.bin.airunner_migrate_knowledge:main",
            "airunner-cleanup-llm-models=airunner.bin.cleanup_llm_models:main",
            "airunner-hf-download=airunner.bin.airunner_hf_download:main",
            "airunner-civitai-download=airunner.bin.airunner_civitai_download:main",
        ],
    },
)



"""Mixin for context menu handling in CustomGraphicsView.

This mixin handles right-click context menus for deleting items.
"""

from PySide6.QtWidgets import QMenu, QGraphicsPixmapItem

from airunner.components.art.gui.widgets.canvas.grid_graphics_item import (
    GridGraphicsItem,
)
from airunner.components.art.gui.widgets.canvas.draggables.active_grid_area import (
    ActiveGridArea,
)


class ContextMenuMixin:
    """Provides context menu functionality for graphics view.

    This mixin manages:
    - Right-click context menu display
    - Item deletion via context menu
    - Text item removal
    - Layer image item removal

    Dependencies:
        - self.mapToScene(): Qt method to map pos to scene
        - self.scene: CustomScene instance
        - self.tr(): Qt translation method
        - self._text_items: List of text items
        - self._remove_text_item(): Text removal method
        - self._remove_layer_image_item(): Layer image removal method
    """

    def contextMenuEvent(self, event):
        """Show delete context menu for items under cursor.

        Displays a context menu with delete action for deletable items
        (excluding grid and active grid area). Handles deletion of text
        items and pixmap items.

        Args:
            event: Qt context menu event.
        """
        try:
            scene_pos = self.mapToScene(event.pos())
        except Exception:
            return

        try:
            items = self.scene.items(scene_pos)
        except Exception:
            items = []

        if not items:
            return

        try:
            # Create context menu with delete action
            menu = QMenu()
            delete_action = menu.addAction(self.tr("Delete"))
            chosen = menu.exec_(event.globalPos())

            if chosen == delete_action:
                # Find first deletable item (skip grid and active grid area)
                deletable = None
                for candidate in items:
                    if not isinstance(
                        candidate, (GridGraphicsItem, ActiveGridArea)
                    ):
                        deletable = candidate
                        break

                if deletable is None:
                    return

                target = deletable

                # Handle text item deletion
                if target in getattr(self, "_text_items", []):
                    self._remove_text_item(target)
                # Handle pixmap item deletion (images)
                elif isinstance(target, QGraphicsPixmapItem):
                    self._remove_layer_image_item(target)
                # Handle other items
                else:
                    getattr(target, "layer_id", None)
                    self._remove_layer_image_item(target)
                    try:
                        if target.scene():
                            target.scene().removeItem(target)
                    except Exception:
                        pass
        except Exception:
            pass



from PIL import Image

from airunner.components.art.filters.base_filter import BaseFilter


class ColorBalanceFilter(BaseFilter):
    """Adjust the color balance of an image.

    This filter allows adjusting the red, green, and blue channels
    independently to achieve color balance corrections.

    Attributes:
        cyan_red: Adjustment factor for the red channel.
        magenta_green: Adjustment factor for the green channel.
        yellow_blue: Adjustment factor for the blue channel.
    """

    def apply_filter(self, image, do_reset=False):
        """Apply color balance adjustments to the image.

        Args:
            image: The PIL Image to filter.
            do_reset: Whether to reset internal state (unused in this filter).

        Returns:
            The color-balanced PIL Image.
        """
        image = image.convert("RGBA")
        red, green, blue, alpha = image.split()

        red = red.point(lambda i: i + (i * self.cyan_red))
        green = green.point(lambda i: i + (i * self.magenta_green))
        blue = blue.point(lambda i: i + (i * self.yellow_blue))

        return Image.merge("RGBA", (red, green, blue, alpha))



from airunner.settings import AIRUNNER_ART_ENABLED


ai_art_models = [
    {
        "name": "Stable Diffusion XL Base 1.0",
        "path": "stabilityai/stable-diffusion-xl-base-1.0",
        "branch": "main",
        "version": "SDXL 1.0",
        "category": "stablediffusion",
        "pipeline_action": "txt2img",
        "enabled": True,
        "model_type": "art",
        "is_default": True,
    },
    {
        "name": "Stable Diffusion XL Turbo",
        "path": "stabilityai/sdxl-turbo",
        "branch": "main",
        "version": "SDXL Turbo",
        "category": "stablediffusion",
        "pipeline_action": "txt2img",
        "enabled": True,
        "model_type": "art",
        "is_default": True,
    },
    {
        "name": "SDXL Inpaint",
        "path": "diffusers/stable-diffusion-xl-1.0-inpainting-0.1",
        "branch": "fp16",
        "version": "SDXL 1.0",
        "category": "stablediffusion",
        "pipeline_action": "inpaint",
        "enabled": True,
        "model_type": "art",
        "is_default": True,
    },
    {
        "name": "Flux.1 S",
        "path": "black-forest-labs/FLUX.1-schnell",
        "branch": "main",
        "version": "Flux.1 S",
        "category": "flux",
        "pipeline_action": "txt2img",
        "enabled": True,
        "model_type": "art",
        "is_default": True,
    },
    {
        "name": "Z-Image Turbo",
        "path": "Tongyi-MAI/Z-Image-Turbo",
        "branch": "main",
        "version": "Z-Image Turbo",
        "category": "zimage",
        "pipeline_action": "txt2img",
        "enabled": True,
        "model_type": "art",
        "is_default": False,
    },
]

llm_models = [
    {
        "name": "Llama 3.1 8B Instruct",
        "path": "meta-llama/Llama-3.1-8B-Instruct",
        "branch": "main",
        "version": "3.1",
        "category": "llm",
        "pipeline_action": "causallm",
        "enabled": True,
        "model_type": "llm",
        "is_default": True,
    },
    {
        "name": "Qwen 2.5 7B Instruct",
        "path": "Qwen/Qwen2.5-7B-Instruct",
        "branch": "main",
        "version": "2.5",
        "category": "llm",
        "pipeline_action": "causallm",
        "enabled": True,
        "model_type": "llm",
        "is_default": False,
    },
    {
        "name": "Command R 8B",
        "path": "CohereForAI/c4ai-command-r-08-2024",
        "branch": "main",
        "version": "08-2024",
        "category": "llm",
        "pipeline_action": "causallm",
        "enabled": True,
        "model_type": "llm",
        "is_default": False,
    },
    {
        "name": "Ministral 8B Instruct",
        "path": "mistralai/Ministral-8B-Instruct-2410",
        "branch": "main",
        "version": "8B-2410",
        "category": "llm",
        "pipeline_action": "causallm",
        "enabled": True,
        "model_type": "llm",
        "is_default": False,
    },
    {
        "name": "Ministral 8B Instruct (Quantized)",
        "path": "w4ffl35/Ministral-8B-Instruct-2410-doublequant",
        "branch": "main",
        "version": "8B-2410-4bit",
        "category": "llm",
        "pipeline_action": "causallm",
        "enabled": True,
        "model_type": "llm",
        "is_default": False,
    },
    {
        "name": "Intfloat E5 Large",
        "path": "intfloat/e5-large",
        "branch": "main",
        "version": "llm",
        "category": "llm",
        "pipeline_action": "embedding",
        "enabled": True,
        "model_type": "llm",
        "is_default": True,
    },
]

if AIRUNNER_ART_ENABLED:
    model_bootstrap_data = ai_art_models + llm_models
else:
    model_bootstrap_data = llm_models



"""Add long-running project tables

Revision ID: 01b52e38f588
Revises: a2b5afa74dde
Create Date: 2025-11-28 04:54:23.299121

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from airunner.utils.db import add_table
from airunner.components.llm.long_running.data.project_state import (
    ProjectState,
    ProjectFeature,
    ProgressEntry,
    SessionState,
    DecisionMemory,
)
from airunner.components.llm.data.fine_tuned_model import FineTunedModel


# revision identifiers, used by Alembic.
revision: str = '01b52e38f588'
down_revision: Union[str, None] = 'a2b5afa74dde'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # Use add_table helper for SQLite compatibility
    # Order matters for foreign key dependencies
    
    # 1. First create tables with no FK dependencies
    add_table(FineTunedModel)
    add_table(ProjectState)  # ProjectState has no circular FK now
    
    # 2. Then create tables that depend on ProjectState
    add_table(ProjectFeature)
    add_table(SessionState)
    add_table(DecisionMemory)
    
    # 3. Finally create tables that depend on multiple tables
    add_table(ProgressEntry)


def downgrade() -> None:
    # Drop tables in reverse order of creation
    op.drop_table('progress_entries')
    op.drop_table('decision_memories')
    op.drop_table('session_states')
    op.drop_table('project_features')
    op.drop_table('project_states')
    op.drop_table('fine_tuned_models')



"""
Dialog for confirming disabling the NSFW filter, with a 'do not show again' checkbox.
"""

from PySide6.QtWidgets import QMessageBox, QCheckBox


def show_nsfw_warning_dialog(parent, show_again_default=True):
    """Show a warning dialog about disabling the NSFW filter.

    Args:
        parent: The parent widget.
        show_again_default: Whether the 'do not show again' box is checked by default.
    Returns:
        (confirmed: bool, do_not_show_again: bool)
    """
    msg_box = QMessageBox(parent)
    msg_box.setWindowTitle("Disable Safety Checker Warning")
    msg_box.setText(
        (
            "WARNING\n\n"
            "You are attempting to disable the safety checker (NSFW filter).\n"
            "It is strongly recommended that you keep this enabled at all times.\n"
            "The Safety Checker prevents potentially harmful content from being displayed.\n"
            "Only disable it if you are sure the Image model you are using is not capable of generating "
            "harmful content.\n"
            "Disabling the safety checker is intended as a last resort for continual false positives and as a "
            "research feature.\n"
            "\n\n"
            "Are you sure you want to disable the filter?"
        )
    )
    msg_box.setStandardButtons(
        QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
    )
    msg_box.setDefaultButton(QMessageBox.StandardButton.No)
    checkbox = QCheckBox("Do not show this warning again")
    checkbox.setChecked(not show_again_default)
    msg_box.setCheckBox(checkbox)
    result = msg_box.exec()
    confirmed = result == QMessageBox.StandardButton.Yes
    do_not_show_again = checkbox.isChecked()
    return confirmed, do_not_show_again



"""Viewport positioning mixin for CustomGraphicsView.

This mixin handles viewport-relative positioning calculations, layer position
management, and viewport compensation for window resizes.
"""

from typing import Dict, Tuple, Optional
from PySide6.QtCore import QPointF

from airunner.components.art.data.canvas_layer import CanvasLayer
from airunner.components.art.data.drawingpad_settings import DrawingPadSettings
from airunner.components.art.utils.canvas_position_manager import (
    CanvasPositionManager,
    ViewState,
)


class ViewportPositioningMixin:
    """Mixin for viewport positioning and layer position management.

    This mixin handles:
    - Viewport center calculations
    - Item recentering within viewport
    - Layer position loading from database
    - Layer position recalculation and saving
    - Canvas items alignment to viewport
    - Active grid area position updates
    - Image position updates based on canvas offset
    - Viewport compensation for window resizes

    Attributes:
        center_pos: Grid origin position (QPointF)
        _grid_compensation_offset: Viewport compensation offset
    """

    @property
    def viewport_center(self) -> QPointF:
        """Calculate the center point of the current viewport.

        Returns:
            Center point of viewport as QPointF.
        """
        viewport_size = self.viewport().size()
        return QPointF(viewport_size.width() / 2, viewport_size.height() / 2)

    def get_recentered_position(
        self, width: float, height: float
    ) -> Tuple[float, float]:
        """Calculate position to center an item of given size in the viewport.

        Args:
            width: Width of item to center.
            height: Height of item to center.

        Returns:
            Tuple of (x, y) coordinates for top-left corner to center the item.
        """
        viewport_center_x = self.viewport_center.x()
        viewport_center_y = self.viewport_center.y()

        item_center_x = width / 2.0
        item_center_y = height / 2.0

        target_x = viewport_center_x - item_center_x
        target_y = viewport_center_y - item_center_y

        return target_x, target_y

    def original_item_positions(self) -> Dict[str, QPointF]:
        """Get the absolute positions of all layer items from the database.

        This method reads saved positions - it does NOT recalculate or modify them.
        Use recenter_layer_positions() to explicitly reposition layers.

        Returns:
            Dictionary mapping scene items to their absolute positions.
        """
        layers = CanvasLayer.objects.order_by("order").all()
        original_item_positions = {}
        for index, layer in enumerate(layers):
            results = DrawingPadSettings.objects.filter_by(layer_id=layer.id)
            if len(results) == 0:
                continue

            drawingpad_settings = results[0]
            scene_item = self.scene._layer_items.get(layer.id)
            if scene_item is None:
                # Layer may not have been materialized yet; skip safely
                continue

            # Read the saved absolute position from the database
            # Do NOT recalculate - preserve what was saved
            if (
                drawingpad_settings.x_pos is not None
                and drawingpad_settings.y_pos is not None
            ):
                pos_x = drawingpad_settings.x_pos
                pos_y = drawingpad_settings.y_pos
            else:
                # If no saved position exists, use center of viewport as fallback
                item_rect = scene_item.boundingRect()
                image_width = item_rect.width()
                image_height = item_rect.height()
                pos_x, pos_y = self.get_recentered_position(
                    int(image_width), int(image_height)
                )
                # Save this calculated position for future use
                DrawingPadSettings.objects.update(
                    drawingpad_settings.id,
                    x_pos=pos_x,
                    y_pos=pos_y,
                )

            original_item_positions[scene_item] = QPointF(pos_x, pos_y)
        return original_item_positions

    def align_canvas_items_to_viewport(self) -> None:
        """Align canvas items (grid and layers) to viewport center.

        Calculates center position if needed, updates grid settings,
        and positions all items appropriately.
        """
        # Don't recalculate center_pos if it was loaded from settings
        # (this preserves the grid origin across app restarts)
        # Only calculate if center_pos is still at the default (0,0)
        if self.center_pos == QPointF(0, 0):
            pos_x, pos_y = self.get_recentered_position(
                self.application_settings.working_width,
                self.application_settings.working_height,
            )
            self.center_pos = QPointF(pos_x, pos_y)
            self.logger.info(
                f"[ALIGN] Center pos calculated: x={pos_x}, y={pos_y}"
            )
        else:
            # Use existing center_pos from loaded settings
            pos_x = int(self.center_pos.x())
            pos_y = int(self.center_pos.y())
            self.logger.info(
                f"[ALIGN] Using loaded center pos: x={pos_x}, y={pos_y}"
            )

        self.update_active_grid_settings(
            pos_x=pos_x,
            pos_y=pos_y,
        )
        # Update display positions
        self.update_active_grid_area_position()

        self.updateImagePositions(self.original_item_positions())

    def update_active_grid_area_position(self) -> None:
        """Update the active grid area position based on current offsets.

        Uses CanvasPositionManager to convert absolute position to display
        position accounting for canvas offset and grid compensation.
        """
        # Skip update during drag to prevent snap-back
        if self.scene and getattr(self.scene, "is_dragging", False):
            self.logger.info(
                "[ACTIVE GRID] Skipping position update - is_dragging is True"
            )
            return

        if self.active_grid_area:
            self.logger.info(
                f"[ACTIVE GRID] update_active_grid_area_position called - current scenePos: {self.active_grid_area.scenePos().x()}, {self.active_grid_area.scenePos().y()}"
            )
            self.logger.info(
                f"[ACTIVE GRID] Settings pos: {self.active_grid_settings.pos_x}, {self.active_grid_settings.pos_y}"
            )

            manager = CanvasPositionManager()
            view_state = ViewState(
                canvas_offset=QPointF(
                    self.canvas_offset_x, self.canvas_offset_y
                ),
                grid_compensation=self._grid_compensation_offset,
            )

            # Convert absolute position to display position
            abs_pos = QPointF(*self.active_grid_settings.pos)
            display_pos = manager.absolute_to_display(abs_pos, view_state)

            self.logger.info(
                f"[ACTIVE GRID] Setting position to display_pos: {display_pos.x()}, {display_pos.y()}"
            )
            self.active_grid_area.setPos(display_pos)

            actual_pos = self.active_grid_area.scenePos()
            self.logger.info(
                f"[ACTIVE GRID] After setPos, actual scenePos: {actual_pos.x()}, {actual_pos.y()}"
            )

    def updateImagePositions(
        self, original_item_positions: Optional[Dict[str, QPointF]] = None
    ) -> None:
        """Update positions of all images in the scene based on canvas offset.

        Args:
            original_item_positions: Optional dict mapping items to absolute positions.
                If None, positions are loaded from database.
        """
        if not self.scene:
            self.logger.error("No scene in updateImagePositions")
            return

        # Use the scene's update_image_position method which handles both
        # the old single-item system and the new layer system
        self.scene.update_image_position(
            self.canvas_offset, original_item_positions
        )

        # Force entire viewport update to handle negative coordinates
        self.viewport().update()
        
    def _apply_viewport_compensation(
        self, shift_x: float, shift_y: float
    ) -> None:
        """Apply viewport center compensation by adjusting grid compensation offset.

        This method shifts only the grid compensation offset so items appear to stay
        centered relative to the viewport, without changing the canvas_offset value
        or the stored absolute positions in the database.

        The CanvasPositionManager will automatically apply the grid_compensation
        when converting absolute positions to display positions.

        Args:
            shift_x: Horizontal shift in viewport center (pixels).
            shift_y: Vertical shift in viewport center (pixels).
        """
        if not self.scene:
            return

        # Skip if the shift is negligible
        if abs(shift_x) < 0.5 and abs(shift_y) < 0.5:
            return

        self.logger.info(
            f"[VIEWPORT COMPENSATION] Applying shift: ({shift_x}, {shift_y}), "
            f"old_compensation=({self._grid_compensation_offset.x()}, {self._grid_compensation_offset.y()}), "
            f"_is_restoring_state={self._is_restoring_state}, _initialized={self._initialized}"
        )

        # Adjust the grid compensation offset
        # This shifts the grid origin to maintain alignment with the viewport center
        self._grid_compensation_offset = QPointF(
            self._grid_compensation_offset.x() + shift_x,
            self._grid_compensation_offset.y() + shift_y,
        )

        self.logger.info(
            f"[VIEWPORT COMPENSATION] New compensation: ({self._grid_compensation_offset.x()}, {self._grid_compensation_offset.y()})"
        )

        # DO NOT modify the scene's original_item_positions here!
        # The CanvasPositionManager.absolute_to_display() already applies
        # grid_compensation in its calculation:
        # display_pos = absolute_pos - canvas_offset + grid_compensation
        #
        # If we shift the absolute positions here, we would be double-applying
        # the compensation, causing items to drift during window resize.

        # Update the visual positions using the new grid_compensation
        # The position manager will automatically apply it
        self.update_active_grid_area_position()
        self.updateImagePositions()



"""Utility functions for NSFW safety checking and image marking."""

import numpy as np
from PIL import Image, ImageDraw, ImageFont
from typing import List, Tuple, Optional, Any


def check_and_mark_nsfw_images(
    images: List[Image.Image],
    feature_extractor: Optional[Any],
    safety_checker: Optional[Any],
    device: str = "cuda",
) -> Tuple[List[Image.Image], List[bool]]:
    """Check images for NSFW content and mark detected images.

    Args:
        images: List of PIL images to check
        feature_extractor: Feature extractor model for preprocessing
        safety_checker: Safety checker model for NSFW detection
        device: Device to run inference on ('cuda' or 'cpu')

    Returns:
        Tuple of (processed_images, nsfw_detected_flags)
        - processed_images: Images with NSFW watermark if detected
        - nsfw_detected_flags: List of booleans indicating NSFW detection
    """
    if not feature_extractor or not safety_checker:
        # If models not loaded, return images unchanged with no detections
        return images, [False] * len(images)

    try:
        # Prepare inputs for safety checker
        safety_checker_input = feature_extractor(
            images, return_tensors="pt"
        ).to(device)

        # Run safety checker
        _, has_nsfw_concepts = safety_checker(
            images=[np.array(img) for img in images],
            clip_input=safety_checker_input.pixel_values.to(device),
        )

        # Mark images with NSFW content
        marked_images = []
        for i, img in enumerate(images):
            if has_nsfw_concepts[i]:
                marked_img = _mark_image_as_nsfw(img)
                marked_images.append(marked_img)
            else:
                marked_images.append(img)

        return marked_images, has_nsfw_concepts

    except Exception as e:
        # On error, return images unchanged
        print(f"Error during NSFW checking: {e}")
        return images, [False] * len(images)


def _mark_image_as_nsfw(image: Image.Image) -> Image.Image:
    """Mark an image as NSFW by blacking it out and adding text overlay.

    Args:
        image: PIL Image to mark

    Returns:
        Marked PIL Image
    """
    # Convert to RGBA for transparency support
    marked = image.convert("RGBA")

    # Black out the entire image
    marked.paste((0, 0, 0), (0, 0, marked.size[0], marked.size[1]))

    # Add "NSFW" text overlay
    draw = ImageDraw.Draw(marked)

    try:
        # Try to load a truetype font
        font = ImageFont.truetype(
            "/usr/share/fonts/truetype/dejavu/DejaVuSans-Bold.ttf", 48
        )
    except Exception:
        # Fall back to default font
        font = ImageFont.load_default()

    # Calculate text position (centered)
    text = "NSFW"
    bbox = draw.textbbox((0, 0), text, font=font)
    text_width = bbox[2] - bbox[0]
    text_height = bbox[3] - bbox[1]

    x = (marked.size[0] - text_width) // 2
    y = (marked.size[1] - text_height) // 2

    # Draw white text
    draw.text((x, y), text, fill=(255, 255, 255, 255), font=font)

    return marked



import unittest

from airunner.components.model_management.quantization_strategy import (
    QuantizationStrategy,
    QuantizationLevel,
)
from airunner.components.model_management.hardware_profiler import (
    HardwareProfile,
)


class TestQuantizationStrategy(unittest.TestCase):
    def setUp(self):
        self.strategy = QuantizationStrategy()
        self.hardware = HardwareProfile(
            total_vram_gb=16.0,
            available_vram_gb=12.0,
            total_ram_gb=32.0,
            available_ram_gb=20.0,
            cuda_available=True,
            cuda_compute_capability=(8, 6),
            device_name="NVIDIA RTX 3080",
            cpu_count=8,
            platform="Linux",
        )

    def test_select_quantization_with_preference(self):
        config = self.strategy.select_quantization(
            model_size_gb=10.0,
            hardware=self.hardware,
            preferred_level=QuantizationLevel.INT4,
        )
        self.assertEqual(config.level, QuantizationLevel.INT4)
        self.assertFalse(config.requires_calibration)

    def test_auto_select_quantization_fp16(self):
        config = self.strategy._auto_select_quantization(
            model_size_gb=7.0, hardware=self.hardware
        )
        self.assertEqual(config.level, QuantizationLevel.FLOAT16)

    def test_auto_select_quantization_int8(self):
        hardware = HardwareProfile(
            total_vram_gb=16.0,
            available_vram_gb=9.0,
            total_ram_gb=32.0,
            available_ram_gb=20.0,
            cuda_available=True,
            cuda_compute_capability=(8, 6),
            device_name="NVIDIA RTX 3080",
            cpu_count=8,
            platform="Linux",
        )
        config = self.strategy._auto_select_quantization(
            model_size_gb=10.0, hardware=hardware
        )
        self.assertEqual(config.level, QuantizationLevel.INT8)

    def test_auto_select_quantization_int4(self):
        hardware = HardwareProfile(
            total_vram_gb=8.0,
            available_vram_gb=6.0,
            total_ram_gb=16.0,
            available_ram_gb=10.0,
            cuda_available=True,
            cuda_compute_capability=(8, 6),
            device_name="NVIDIA RTX 3060",
            cpu_count=8,
            platform="Linux",
        )
        config = self.strategy._auto_select_quantization(
            model_size_gb=10.0, hardware=hardware
        )
        self.assertEqual(config.level, QuantizationLevel.INT4)

    def test_get_config_for_level_fp16(self):
        config = self.strategy._get_config_for_level(
            QuantizationLevel.FLOAT16, 10.0
        )
        self.assertEqual(config.level, QuantizationLevel.FLOAT16)
        self.assertEqual(config.estimated_memory_gb, 20.0)
        self.assertFalse(config.requires_calibration)

    def test_get_config_for_level_int4(self):
        config = self.strategy._get_config_for_level(
            QuantizationLevel.INT4, 10.0
        )
        self.assertEqual(config.level, QuantizationLevel.INT4)
        self.assertEqual(config.estimated_memory_gb, 5.0)
        self.assertFalse(config.requires_calibration)

    def test_get_config_for_level_int2(self):
        config = self.strategy._get_config_for_level(
            QuantizationLevel.INT2, 10.0
        )
        self.assertEqual(config.level, QuantizationLevel.INT2)
        self.assertEqual(config.estimated_memory_gb, 2.5)
        self.assertTrue(config.requires_calibration)


if __name__ == "__main__":
    unittest.main()



"""
Research document management tools.

Provides tools for creating, locking, and managing research documents
during the Deep Research workflow. Documents are created in the document
editor and locked to prevent user editing while the LLM is working.
"""

import os
from datetime import datetime
from typing import Annotated, Any

from airunner.components.llm.core.tool_registry import tool, ToolCategory
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger
from airunner.enums import SignalCode

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)

# Research document template
RESEARCH_DOCUMENT_TEMPLATE = """# {title}

**Research Date:** {date}  
**Research Started:** {datetime_started}  
**Status:** In Progress

---
"""


@tool(
    name="create_research_document",
    category=ToolCategory.RESEARCH,
    description=(
        "Create a new research document in the document editor. "
        "Returns the file path. The document will be locked for editing by the LLM. "
        "Use this at the start of deep research to create the main research document."
    ),
    return_direct=False,
    requires_api=True,
)
def create_research_document(
    topic: Annotated[str, "The research topic/title"],
    api: Any = None,
) -> str:
    """
    Create a new research document.

    Creates a markdown file in text/other/research/ with a standardized
    academic template. The document is automatically opened in the editor
    and locked to prevent user modifications.

    Args:
        topic: Research topic/title
        api: API instance (injected automatically)

    """
    if not api:
        return "Error: API not available"

    try:
        # Get base path from settings
        base_path = api.path_settings.base_path
        research_dir = os.path.join(base_path, "text", "other", "research")
        os.makedirs(research_dir, exist_ok=True)

        # Generate filename with timestamp
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        safe_topic = "".join(
            c for c in topic if c.isalnum() or c in (" ", "_", "-")
        ).replace(" ", "_")[:50]
        filename = f"{safe_topic}_{timestamp}.md"
        filepath = os.path.join(research_dir, filename)

        # Create document from template
        content = RESEARCH_DOCUMENT_TEMPLATE.format(
            title=topic,
            date=datetime.now().strftime("%B %d, %Y"),
            datetime_started=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        )

        # Write the file
        with open(filepath, "w", encoding="utf-8") as f:
            f.write(content)

        logger.info(f"Created research document: {filepath}")

        # Signal to open document in editor (locked)
        api.emit_signal(
            SignalCode.OPEN_RESEARCH_DOCUMENT,
            {"path": filepath, "locked": True, "title": topic},
        )

        return filepath

    except Exception as e:
        logger.error(f"Failed to create research document: {e}", exc_info=True)
        return f"Error creating research document: {str(e)}"


@tool(
    name="create_research_notes",
    category=ToolCategory.RESEARCH,
    description=(
        "Create a temporary notes file for incremental research notes. "
        "This file stores findings from web searches organized by URL. "
        "Use this to accumulate research before writing the main document."
    ),
    return_direct=False,
    requires_api=True,
)
def create_research_notes(
    topic: Annotated[str, "The research topic (should match main document)"],
    api: Any = None,
) -> str:
    """
    Create a research notes file.

    Creates a markdown file for temporary notes during research.
    Notes are organized by source URL.

    Args:
        topic: Research topic (should match main document)
        api: API instance (injected automatically)

    """
    if not api:
        return "Error: API not available"

    try:
        base_path = api.path_settings.base_path
        research_dir = os.path.join(base_path, "text", "other", "research")
        os.makedirs(research_dir, exist_ok=True)

        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        safe_topic = "".join(
            c for c in topic if c.isalnum() or c in (" ", "_", "-")
        ).replace(" ", "_")[:50]
        filename = f"{safe_topic}_{timestamp}_notes.md"
        filepath = os.path.join(research_dir, filename)

        # Create notes file
        content = f"""# Research Notes: {topic}

**Date:** {datetime.now().strftime("%B %d, %Y %H:%M")}

---

## Research Sources

"""
        with open(filepath, "w", encoding="utf-8") as f:
            f.write(content)

        logger.info(f"Created research notes: {filepath}")

        # Signal to open notes in editor (read-only/locked during research)
        if api and hasattr(api, "emit_signal"):
            api.emit_signal(
                SignalCode.OPEN_RESEARCH_DOCUMENT,
                {"path": filepath, "locked": True, "title": f"Notes: {topic}"},
            )

        return filepath

    except Exception as e:
        logger.error(f"Failed to create notes file: {e}", exc_info=True)
        return f"Error creating notes file: {str(e)}"


@tool(
    name="append_research_notes",
    category=ToolCategory.RESEARCH,
    description=(
        "Append findings from a source to the research notes file. "
        "Organizes notes by URL with header and content. "
        "Use this after scraping each web page during research phase."
    ),
    return_direct=False,
    requires_api=True,
)
def append_research_notes(
    notes_path: Annotated[str, "Path to the notes file"],
    source_url: Annotated[str, "URL of the source"],
    findings: Annotated[str, "Key findings/summary from this source"],
    api: Any = None,
) -> str:
    """
    Append findings to research notes.

    Args:
        notes_path: Path to notes file
        source_url: URL of source
        findings: Summary of findings

    """
    try:
        # Build the entry
        entry = f"""
### {source_url}

**Retrieved:** {datetime.now().strftime("%Y-%m-%d %H:%M")}

{findings}

---

"""

        # Append to file
        with open(notes_path, "a", encoding="utf-8") as f:
            f.write(entry)

        # Stream the content to open document (if any)
        if api:
            # Send the full entry as a streaming update
            api.emit_signal(
                SignalCode.STREAM_TO_DOCUMENT,
                {"path": notes_path, "chunk": entry},
            )

        logger.info(f"Appended notes from {source_url}")
        return f"Successfully added notes from {source_url}"

    except Exception as e:
        logger.error(f"Failed to append notes: {e}", exc_info=True)
        return f"Error appending notes: {str(e)}"


@tool(
    name="update_research_section",
    category=ToolCategory.RESEARCH,
    description=(
        "Update a specific section of the research document. "
        "Can update Abstract, Introduction, body sections, Conclusion, or Sources. "
        "Use this to incrementally build the research document."
    ),
    return_direct=False,
    requires_api=True,
)
def update_research_section(
    document_path: Annotated[str, "Path to the research document"],
    section_name: Annotated[
        str,
        "Section to update (e.g., 'Abstract', 'Introduction', 'Conclusion')",
    ],
    content: Annotated[str, "New content for the section"],
    api: Any = None,
) -> str:
    """
    Update a section of the research document.

    Args:
        document_path: Path to research document
        section_name: Name of section to update
        content: New content
        api: API instance (injected automatically)

    """
    try:
        # Read current document
        with open(document_path, "r", encoding="utf-8") as f:
            doc_content = f.read()

        # Find and replace section
        # Look for "## {section_name}" and replace until next "##" or "---"
        import re

        # Pattern to match section header and content until next section
        # Matches: optional whitespace, "## Section", newline, content until next "##" or "---" or end
        pattern = rf"^[ \t]*## {re.escape(section_name)}\s*\n(.*?)(?=^[ \t]*##|^---|\Z)"

        def replace_section(match):
            # Return just the header and new content (preserve proper spacing)
            return f"## {section_name}\n\n{content}\n\n"

        new_content, count = re.subn(
            pattern,
            replace_section,
            doc_content,
            flags=re.DOTALL | re.MULTILINE,
        )

        if count == 0:
            # Section not found, append it before final "---" or at end
            # Try to insert before the final document separator
            if "\n---\n" in doc_content:
                # Insert before the last "---"
                parts = doc_content.rsplit("\n---\n", 1)
                new_content = f"{parts[0]}\n\n## {section_name}\n\n{content}\n\n---\n{parts[1]}"
            else:
                # No separator, just append
                new_content = (
                    f"{doc_content}\n\n## {section_name}\n\n{content}\n"
                )
            logger.info(f"Added new section: {section_name}")
        else:
            logger.info(
                f"Updated section: {section_name} (replaced {count} occurrence(s))"
            )

        # Write updated document
        with open(document_path, "w", encoding="utf-8") as f:
            f.write(new_content)

        # Signal to update the document in memory
        if api:
            api.emit_signal(
                SignalCode.UPDATE_DOCUMENT_CONTENT,
                {
                    "path": document_path,
                    "content": new_content,
                    "append": False,
                },
            )
            logger.info(f"Signaled document update for: {document_path}")

        return f"Successfully updated {section_name} section"

    except Exception as e:
        logger.error(f"Failed to update section: {e}", exc_info=True)
        return f"Error updating section: {str(e)}"


@tool(
    name="add_source_citation",
    category=ToolCategory.RESEARCH,
    description=(
        "Add a source citation to the Sources section of the research document. "
        "Formats citations with title, URL, and access date. "
        "Use this to build the bibliography as research progresses."
    ),
    return_direct=False,
    requires_api=False,
)
def add_source_citation(
    document_path: Annotated[str, "Path to the research document"],
    title: Annotated[str, "Title of the source"],
    url: Annotated[str, "URL of the source"],
    author: Annotated[
        str, "Author (if known, otherwise 'Unknown')"
    ] = "Unknown",
) -> str:
    """
    Add a citation to the Sources section.

    Args:
        document_path: Path to research document
        title: Source title
        url: Source URL
        author: Author name

    """
    try:
        # Read current document
        with open(document_path, "r", encoding="utf-8") as f:
            content = f.read()

        # Find Sources section
        import re

        access_date = datetime.now().strftime("%B %d, %Y")

        citation = f"""
- **{title}**  
  Author: {author}  
  URL: [{url}]({url})  
  Accessed: {access_date}
"""

        # Find Sources section and append
        pattern = r"(## Sources\s*\n)(.*?)(?=\n##|\Z)"

        def add_citation(match):
            existing = match.group(2).strip()
            if "[To be added" in existing or not existing:
                # Replace placeholder
                return f"{match.group(1)}\n{citation}"
            else:
                # Append to existing
                return f"{match.group(1)}{existing}\n{citation}"

        new_content, count = re.subn(
            pattern, add_citation, content, flags=re.DOTALL
        )

        if count == 0:
            # Sources section not found, add it
            new_content = content + f"\n\n## Sources\n{citation}\n"

        with open(document_path, "w", encoding="utf-8") as f:
            f.write(new_content)

        logger.info(f"Added citation: {title}")
        return f"Successfully added citation for {title}"

    except Exception as e:
        logger.error(f"Failed to add citation: {e}", exc_info=True)
        return f"Error adding citation: {str(e)}"


@tool(
    name="finalize_research_document",
    category=ToolCategory.RESEARCH,
    description=(
        "Mark research document as complete and unlock it for user viewing/editing. "
        "Updates status and signals the editor to unlock the document."
    ),
    return_direct=False,
    requires_api=True,
)
def finalize_research_document(
    document_path: Annotated[str, "Path to the research document"],
    api: Any = None,
) -> str:
    """
    Finalize and unlock research document.

    Args:
        document_path: Path to research document
        api: API instance

    """
    if not api:
        return "Error: API not available"

    try:
        # Update status in document
        with open(document_path, "r", encoding="utf-8") as f:
            content = f.read()

        content = content.replace(
            "**Status:** In Progress", "**Status:** Complete"
        )

        with open(document_path, "w", encoding="utf-8") as f:
            f.write(content)

        # Signal to unlock document
        api.emit_signal(
            SignalCode.UNLOCK_RESEARCH_DOCUMENT,
            {"path": document_path},
        )

        logger.info(f"Finalized research document: {document_path}")
        return f"Research document completed and unlocked: {document_path}"

    except Exception as e:
        logger.error(f"Failed to finalize document: {e}", exc_info=True)
        return f"Error finalizing document: {str(e)}"


@tool(
    name="load_indexed_documents_into_rag",
    category=ToolCategory.RESEARCH,
    description=(
        "Search for and load indexed documents related to a research topic "
        "into RAG context. Use this BEFORE starting web research to incorporate "
        "existing knowledge from the local knowledge base."
    ),
    return_direct=False,
    requires_api=True,
)
def load_indexed_documents_into_rag(
    query: Annotated[
        str,
        "Research topic or query to find relevant indexed documents",
    ],
    max_documents: Annotated[
        int, "Maximum number of documents to load (default 5)"
    ] = 5,
    api: Any = None,
) -> str:
    """Search indexed documents and load relevant ones into RAG context.

    This tool combines document discovery and loading in one step. It:
    1. Searches the knowledge base for documents matching the query
    2. Filters for indexed documents only
    3. Loads them into RAG context for use in research

    Use this at the beginning of research to incorporate existing knowledge
    from your local knowledge base before searching the web.

    Args:
        query: Research topic or query to find relevant documents
        max_documents: Maximum number of documents to load (default 5)
        api: API instance (injected)


    Usage:
        load_indexed_documents_into_rag("machine learning transformers", 3)
    """
    try:
        from airunner.components.documents.data.models.document import Document
        from airunner.components.data.session_manager import session_scope

        logger.info(f"Loading indexed documents for query: {query}")

        # Search for indexed documents
        with session_scope() as session:
            docs = (
                session.query(Document)
                .filter_by(active=True, indexed=True)
                .all()
            )

            if not docs:
                return (
                    "No indexed documents found in knowledge base. "
                    "Proceeding with web-only research."
                )

            # Score documents by relevance
            query_lower = query.lower()
            query_terms = query_lower.split()

            scored_docs = []
            for doc in docs:
                path_lower = doc.path.lower()
                filename = os.path.basename(path_lower)

                score = 0
                for term in query_terms:
                    if term in filename:
                        score += 10
                    elif term in path_lower:
                        score += 5

                if score > 0:
                    scored_docs.append((score, doc))

            # Sort and take top N
            scored_docs.sort(reverse=True, key=lambda x: x[0])
            top_docs = scored_docs[:max_documents]

            if not top_docs:
                return (
                    f"No indexed documents match '{query}'. "
                    f"Proceeding with web-only research."
                )

            # Load documents into RAG
            rag_manager = getattr(api, "rag_manager", None) if api else None

            if not rag_manager:
                # RAG not available, return document paths for reference
                result_parts = [
                    f"Found {len(top_docs)} relevant indexed document(s), "
                    f"but RAG manager not available:\n"
                ]
                for i, (score, doc) in enumerate(top_docs, 1):
                    filename = os.path.basename(doc.path)
                    result_parts.append(f"{i}. {filename}")
                    result_parts.append(f"   Path: {doc.path}")
                return "\n".join(result_parts)

            # Actually load into RAG and search
            loaded_docs = []
            excerpts = []

            for score, doc in top_docs:
                try:
                    # Check if document is already in RAG by searching
                    # If not, we'd need to trigger indexing via signal
                    # For now, just search RAG with the query
                    results = rag_manager.search(query, k=2)

                    if results:
                        for result in results:
                            source = result.metadata.get("source", "unknown")
                            if source == doc.path:
                                content = (
                                    result.page_content[:300]
                                    if len(result.page_content) > 300
                                    else result.page_content
                                )
                                excerpts.append(
                                    f"[{os.path.basename(doc.path)}]\n{content}"
                                )
                                loaded_docs.append(os.path.basename(doc.path))
                                break
                except Exception as e:
                    logger.warning(f"Failed to search RAG for {doc.path}: {e}")
                    continue

            # Format response
            if loaded_docs:
                result_parts = [
                    f"Loaded {len(loaded_docs)} indexed document(s) into "
                    f"RAG context for '{query}':\n"
                ]
                for doc_name in loaded_docs:
                    result_parts.append(f"- {doc_name}")

                if excerpts:
                    result_parts.append("\nRelevant excerpts:")
                    result_parts.extend(excerpts)

                result_parts.append(
                    "\nThese documents are now available for rag_search. "
                    "Use rag_search to find specific information from them."
                )
                return "\n\n".join(result_parts)
            else:
                # Documents found but not loaded in RAG yet
                result_parts = [
                    f"Found {len(top_docs)} relevant indexed document(s) "
                    f"for '{query}':\n"
                ]
                for i, (score, doc) in enumerate(top_docs, 1):
                    filename = os.path.basename(doc.path)
                    result_parts.append(f"{i}. {filename}")

                result_parts.append(
                    "\nNote: These documents are indexed but may not be "
                    "loaded into RAG yet. Proceeding with web research."
                )
                return "\n".join(result_parts)

    except Exception as e:
        logger.error(f"Error loading indexed documents: {e}", exc_info=True)
        return (
            f"Error loading indexed documents: {str(e)}. "
            f"Proceeding with web-only research."
        )


@tool(
    name="edit_document_find_replace",
    category=ToolCategory.RESEARCH,
    description=(
        "Edit a document by finding and replacing text. "
        "Supports simple string replacement and regular expressions. "
        "Use this for precise edits, correcting errors, or updating specific parts of the document."
    ),
    return_direct=False,
    requires_api=True,
)
def edit_document_find_replace(
    document_path: Annotated[str, "Path to the document to edit"],
    find_text: Annotated[str, "Text or pattern to find"],
    replace_text: Annotated[str, "Text to replace with"],
    is_regex: Annotated[
        bool, "Whether find_text is a regular expression"
    ] = False,
    api: Any = None,
) -> str:
    """
    Edit a document using find and replace.

    Args:
        document_path: Path to the document
        find_text: Text to find
        replace_text: Replacement text
        is_regex: Treat find_text as regex
        api: API instance

    """
    import re

    try:
        if not os.path.exists(document_path):
            return f"Error: Document not found at {document_path}"

        # Read current document
        with open(document_path, "r", encoding="utf-8") as f:
            content = f.read()

        # Perform replacement
        if is_regex:
            new_content, count = re.subn(find_text, replace_text, content)
        else:
            count = content.count(find_text)
            new_content = content.replace(find_text, replace_text)

        if count == 0:
            return f"Text not found: '{find_text}'"

        # Write updated document
        with open(document_path, "w", encoding="utf-8") as f:
            f.write(new_content)

        logger.info(
            f"Updated document {document_path}: replaced {count} occurrence(s) of '{find_text}'"
        )

        # Signal to update the document in memory
        if api:
            api.emit_signal(
                SignalCode.UPDATE_DOCUMENT_CONTENT,
                {
                    "path": document_path,
                    "content": new_content,
                    "append": False,
                },
            )
            logger.info(f"Signaled document update for: {document_path}")

        return f"Successfully replaced {count} occurrence(s) of text."

    except Exception as e:
        logger.error(f"Failed to edit document: {e}", exc_info=True)
        return f"Error editing document: {str(e)}"



"""
Tests for code_validator.py

Test validation of Python code with flake8 and mypy.
"""

import pytest

from airunner.components.llm.tools.code_validator import (
    CodeValidator,
    ValidationIssue,
    ValidationResult,
    ValidationSeverity,
)


@pytest.fixture
def temp_py_file(tmp_path):
    """Create a temporary Python file."""

    def _create(content: str, filename: str = "test.py"):
        file_path = tmp_path / filename
        file_path.write_text(content)
        return str(file_path)

    return _create


@pytest.fixture
def validator():
    """Create CodeValidator instance."""
    return CodeValidator(enable_flake8=True, enable_mypy=True)


def test_validator_initialization(validator):
    """Test validator initializes correctly."""
    assert isinstance(validator, CodeValidator)
    # Tools may or may not be available depending on environment
    assert validator.enable_flake8 in [True, False]
    assert validator.enable_mypy in [True, False]


def test_validate_nonexistent_file(validator):
    """Test validation of non-existent file."""
    result = validator.validate_file("/nonexistent/file.py")

    assert result.success is False
    assert result.error_count == 0
    assert len(result.issues) == 0


def test_validate_valid_python_file(validator, temp_py_file):
    """Test validation of valid Python file."""
    content = '''
def hello(name: str) -> str:
    """Say hello."""
    return f"Hello, {name}!"
'''

    file_path = temp_py_file(content)
    result = validator.validate_file(file_path)

    # Should have no errors (may have warnings about unused import etc.)
    assert result.file_path == file_path
    assert isinstance(result, ValidationResult)


def test_validate_file_with_syntax_error(validator, temp_py_file):
    """Test validation of file with syntax error."""
    content = """
def hello(name: str -> str:  # Missing closing paren
    return f"Hello, {name}!"
"""

    file_path = temp_py_file(content)
    result = validator.validate_file(file_path)

    # Should detect syntax error if flake8 is available
    if validator.enable_flake8:
        assert result.error_count > 0 or result.has_errors


def test_validate_file_with_style_issues(validator, temp_py_file):
    """Test validation of file with PEP 8 style issues."""
    content = """
def hello( name ):  # Extra spaces
    x=1+2  # Missing spaces around operators
    return "Hello, " + name
"""

    file_path = temp_py_file(content)
    result = validator.validate_file(file_path)

    # Should detect style issues if flake8 is available
    if validator.enable_flake8:
        assert len(result.issues) > 0


def test_validate_file_with_type_errors(validator, temp_py_file):
    """Test validation of file with type errors."""
    content = """
def add(a: int, b: int) -> int:
    return a + b

result: int = add("1", "2")  # Type error: strings passed for ints
"""

    file_path = temp_py_file(content)
    validator.validate_file(file_path)

    # Should detect type errors if mypy is available
    if validator.enable_mypy:
        # mypy should find the type error
        pass  # Result depends on mypy availability


def test_validation_result_properties(temp_py_file):
    """Test ValidationResult properties."""
    file_path = temp_py_file("x = 1\n")

    result = ValidationResult(
        file_path=file_path,
        success=False,
        issues=[
            ValidationIssue(
                file_path=file_path,
                line=1,
                column=1,
                severity=ValidationSeverity.ERROR,
                code="E501",
                message="line too long",
                tool="flake8",
            ),
            ValidationIssue(
                file_path=file_path,
                line=2,
                column=1,
                severity=ValidationSeverity.WARNING,
                code="W291",
                message="trailing whitespace",
                tool="flake8",
            ),
        ],
        error_count=1,
        warning_count=1,
        info_count=0,
    )

    assert result.has_errors is True
    assert result.has_warnings is True
    assert len(result.issues) == 2
    assert "1 errors, 1 warnings" in str(result)


def test_validation_issue_str():
    """Test ValidationIssue string representation."""
    issue = ValidationIssue(
        file_path="/path/test.py",
        line=10,
        column=5,
        severity=ValidationSeverity.ERROR,
        code="E501",
        message="line too long",
        tool="flake8",
    )

    issue_str = str(issue)
    assert "/path/test.py" in issue_str
    assert ":10:5:" in issue_str
    assert "error" in issue_str
    assert "E501" in issue_str
    assert "line too long" in issue_str


def test_parse_flake8_output(validator):
    """Test parsing flake8 output."""
    output = """test.py:10:5: E501 line too long (82 > 79 characters)
test.py:15:1: W291 trailing whitespace
test.py:20:10: E999 SyntaxError: invalid syntax"""

    issues = validator._parse_flake8_output(output, "test.py")

    assert len(issues) == 3

    assert issues[0].line == 10
    assert issues[0].column == 5
    assert issues[0].code == "E501"
    assert issues[0].severity == ValidationSeverity.ERROR
    assert issues[0].tool == "flake8"

    assert issues[1].code == "W291"
    assert issues[1].severity == ValidationSeverity.WARNING

    assert issues[2].code == "E999"


def test_parse_mypy_output(validator):
    """Test parsing mypy output."""
    output = """test.py:10:5: error: Incompatible return value type [return-value]
test.py:15:10: warning: Unused variable "x" [unused-variable]
test.py:20:1: note: In function "hello" [note]"""

    issues = validator._parse_mypy_output(output, "test.py")

    assert len(issues) == 3

    assert issues[0].line == 10
    assert issues[0].column == 5
    assert issues[0].code == "return-value"
    assert issues[0].severity == ValidationSeverity.ERROR
    assert issues[0].tool == "mypy"

    assert issues[1].severity == ValidationSeverity.WARNING
    assert issues[1].code == "unused-variable"

    assert issues[2].severity == ValidationSeverity.INFO


def test_validate_multiple_files(validator, temp_py_file):
    """Test validating multiple files."""
    file1 = temp_py_file("x = 1\n", "file1.py")
    file2 = temp_py_file("y = 2\n", "file2.py")

    results = validator.validate_files([file1, file2])

    assert len(results) == 2
    assert all(isinstance(r, ValidationResult) for r in results)


def test_get_summary(validator, temp_py_file):
    """Test getting validation summary."""
    file1 = temp_py_file("x = 1\n", "file1.py")

    results = [
        ValidationResult(
            file_path=file1,
            success=False,
            issues=[],
            error_count=2,
            warning_count=1,
            info_count=0,
        ),
        ValidationResult(
            file_path=file1,
            success=True,
            issues=[],
            error_count=0,
            warning_count=3,
            info_count=0,
        ),
    ]

    summary = validator.get_summary(results)

    assert summary["total_files"] == 2
    assert summary["files_with_errors"] == 1
    assert summary["files_with_warnings"] == 2
    assert summary["total_errors"] == 2
    assert summary["total_warnings"] == 4
    assert summary["success"] is False


def test_validator_with_disabled_tools():
    """Test validator with all tools disabled."""
    validator = CodeValidator(enable_flake8=False, enable_mypy=False)

    assert validator.enable_flake8 is False
    assert validator.enable_mypy is False


def test_parse_flake8_output_no_matches(validator):
    """Test parsing flake8 output with no matches."""
    output = "Some random text\nNot a valid flake8 line"

    issues = validator._parse_flake8_output(output, "test.py")

    assert len(issues) == 0


def test_parse_mypy_output_without_code(validator):
    """Test parsing mypy output without error code."""
    output = "test.py:10:5: error: Some error message"

    issues = validator._parse_mypy_output(output, "test.py")

    assert len(issues) == 1
    assert issues[0].code == "mypy"  # Default code when not specified


def test_validation_result_success_with_warnings(temp_py_file):
    """Test that ValidationResult is success even with warnings."""
    file_path = temp_py_file("x = 1\n")

    result = ValidationResult(
        file_path=file_path,
        success=True,
        issues=[
            ValidationIssue(
                file_path=file_path,
                line=1,
                column=1,
                severity=ValidationSeverity.WARNING,
                code="W291",
                message="trailing whitespace",
                tool="flake8",
            ),
        ],
        error_count=0,
        warning_count=1,
        info_count=0,
    )

    assert result.success is True
    assert result.has_errors is False
    assert result.has_warnings is True



"""Mixins for NodeGraphWorker to reduce class complexity."""

from airunner.components.nodegraph.workers.mixins.workflow_control_mixin import (
    WorkflowControlMixin,
)
from airunner.components.nodegraph.workers.mixins.message_handling_mixin import (
    MessageHandlingMixin,
)
from airunner.components.nodegraph.workers.mixins.queue_processing_mixin import (
    QueueProcessingMixin,
)
from airunner.components.nodegraph.workers.mixins.node_execution_mixin import (
    NodeExecutionMixin,
)

__all__ = [
    "WorkflowControlMixin",
    "MessageHandlingMixin",
    "QueueProcessingMixin",
    "NodeExecutionMixin",
]



"""
Example eval tests demonstrating proper architecture.

These tests show how to:
1. Test real LLM behavior (not mocked)
2. Mock only external side effects
3. Use quality metrics to evaluate responses
4. Test tool usage patterns

Run with:
    pytest test_eval_example.py -v
    pytest test_eval_example.py --model qwen2.5-coder:32b
"""

import pytest
import logging
from unittest.mock import patch, Mock, MagicMock
from airunner.components.eval.utils.tracking import track_trajectory_sync
from airunner.components.eval.utils.quality_metrics import (
    evaluate_tool_usage,
    evaluate_response_quality,
    assert_quality_threshold,
)

logger = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.eval,
    pytest.mark.timeout(60),
]


@pytest.mark.eval
class TestEvalExample:
    """Example eval tests showing proper architecture."""

    @patch("airunner.components.llm.tools.agent_tools.AgentConfig")
    @patch("airunner.components.llm.tools.agent_tools.session_scope")
    def test_agent_creation_with_quality_metrics(
        self,
        mock_session_scope,
        mock_agent_config,
        airunner_client_function_scope,
    ):
        """
        Test agent creation with quality metrics.

        This test demonstrates proper eval test architecture:
        1. Mocks database operations (side effects)
        2. Uses real LLM (tests intelligence)
        3. Evaluates tool usage quality
        4. Evaluates response quality
        """
        # Mock database operations (side effect, not intelligence)
        mock_session = MagicMock()
        mock_session_scope.return_value.__enter__.return_value = mock_session

        mock_agent = Mock()
        mock_agent.id = 1
        mock_agent.name = "Helper"
        mock_agent_config.return_value = mock_agent

        prompt = "Create a new agent named Helper to assist with coding tasks"

        # Real LLM call - tests understanding and tool selection
        result = track_trajectory_sync(
            airunner_client_function_scope,
            prompt=prompt,
            max_tokens=500,
            tool_categories=["AGENT"],
        )

        # Evaluate tool usage quality
        tool_metrics = evaluate_tool_usage(
            result,
            expected_tools=["create_agent"],
            max_calls=2,  # Should not call more than twice
        )

        logger.info(f"Tool usage metrics: {tool_metrics}")

        # Assert tool usage quality
        assert tool_metrics["tool_called"], "No tools were called"
        assert tool_metrics[
            "correct_tools"
        ], f"Wrong tools called: {tool_metrics['details']}"
        assert not tool_metrics["overcalling"], "LLM is over-calling tools"
        assert_quality_threshold(tool_metrics, 0.7, "tool usage")

        # Evaluate response quality
        quality_metrics = evaluate_response_quality(
            result["response"],
            prompt,
            expected_keywords=["agent", "Helper", "created"],
            forbidden_keywords=[
                "error",
                "failed",
                "cannot",
            ],  # Hallucination check
        )

        logger.info(f"Response quality metrics: {quality_metrics}")

        # Assert response quality
        assert (
            quality_metrics["relevance"] > 0.6
        ), f"Response not relevant: {quality_metrics['details']}"
        assert quality_metrics[
            "no_hallucination"
        ], "Response contains hallucinated errors"
        assert_quality_threshold(quality_metrics, 0.7, "response quality")

    @patch("airunner.components.llm.tools.web_tools.AggregatedSearchTool")
    def test_web_search_tool_selection(
        self,
        mock_search_tool,
        airunner_client_function_scope,
    ):
        """
        Test LLM correctly triggers web search for information requests.

        Demonstrates:
        - Mocking external HTTP requests (side effect)
        - Testing real LLM tool selection (intelligence)
        - Verifying tool was called with correct arguments
        """
        # Mock external HTTP request (side effect, not intelligence)
        mock_search_tool.aggregated_search_sync.return_value = {
            "duckduckgo": [
                {
                    "title": "Python Tutorial - Official Docs",
                    "link": "https://docs.python.org/3/tutorial/",
                    "snippet": "The Python Tutorial — Python 3 documentation",
                }
            ]
        }

        prompt = "Search for Python tutorials and summarize what you find"

        # Real LLM call - tests when to use search
        result = track_trajectory_sync(
            airunner_client_function_scope,
            prompt=prompt,
            max_tokens=500,
            tool_categories=["SEARCH"],
        )

        # Evaluate tool usage
        tool_metrics = evaluate_tool_usage(
            result,
            expected_tools=["search_web"],  # Or similar tool name
            max_calls=1,  # Should only search once
        )

        logger.info(f"Tool metrics: {tool_metrics}")

        # Verify LLM understood it needs to search
        assert tool_metrics[
            "tool_called"
        ], "LLM should have triggered search tool for information request"

        # Verify mock was called (tool was actually invoked)
        if mock_search_tool.aggregated_search_sync.called:
            call_args = mock_search_tool.aggregated_search_sync.call_args
            query = call_args[0][0].lower()

            # Verify search query contains relevant terms
            assert "python" in query, "Search query missing 'python'"
            assert (
                "tutorial" in query or "tutorials" in query
            ), "Search query missing 'tutorial'"

        # Evaluate response quality
        quality_metrics = evaluate_response_quality(
            result["response"],
            prompt,
            expected_keywords=["python", "tutorial"],
            min_length=50,  # Should summarize findings
        )

        logger.info(f"Response quality: {quality_metrics}")

        # Response should mention search results
        assert (
            quality_metrics["completeness"] > 0.0
        ), "Response too short - should summarize search results"

    @patch("airunner.components.llm.tools.agent_tools.session_scope")
    def test_tool_avoidance_when_not_needed(
        self,
        mock_session_scope,
        airunner_client_function_scope,
    ):
        """
        Test LLM does NOT call tools when they're not needed.

        This tests that the LLM:
        - Understands when tools are NOT necessary
        - Avoids over-calling tools
        - Can answer directly when appropriate
        """
        mock_session = MagicMock()
        mock_session_scope.return_value.__enter__.return_value = mock_session

        # Simple factual question - no tools needed
        prompt = "What is 2 + 2?"

        result = track_trajectory_sync(
            airunner_client_function_scope,
            prompt=prompt,
            max_tokens=100,
            tool_categories=[
                "AGENT",
                "SEARCH",
            ],  # Tools available but not needed
        )

        # Evaluate tool usage
        tool_metrics = evaluate_tool_usage(
            result,
            expected_tools=[],  # No tools should be called
            max_calls=0,
        )

        logger.info(f"Tool avoidance metrics: {tool_metrics}")

        # LLM should NOT have called tools
        assert len(result["tools"]) == 0 or tool_metrics["overcalling"], (
            f"LLM unnecessarily called tools: {result['tools']}\n"
            f"Should answer simple math directly without tools"
        )

        # Response should still be correct
        quality_metrics = evaluate_response_quality(
            result["response"],
            prompt,
            expected_keywords=["4"],  # Correct answer
            min_length=1,
        )

        assert (
            quality_metrics["relevance"] > 0.5
        ), "Response should contain the answer '4'"

    @patch("airunner.components.llm.tools.agent_tools.AgentConfig")
    @patch("airunner.components.llm.tools.agent_tools.session_scope")
    def test_tool_argument_quality(
        self,
        mock_session_scope,
        mock_agent_config,
        airunner_client_function_scope,
    ):
        """
        Test LLM passes correct arguments to tools.

        This verifies the LLM:
        - Extracts correct parameters from user prompt
        - Passes them to tools in proper format
        - Doesn't hallucinate or fabricate arguments
        """
        mock_session = MagicMock()
        mock_session_scope.return_value.__enter__.return_value = mock_session

        mock_agent = Mock()
        mock_agent.id = 1
        mock_agent.name = "CodeHelper"
        mock_agent_config.return_value = mock_agent

        prompt = (
            "Create an agent named CodeHelper with description "
            "'Assists with Python programming'"
        )

        result = track_trajectory_sync(
            airunner_client_function_scope,
            prompt=prompt,
            max_tokens=500,
            tool_categories=["AGENT"],
        )

        # Find the create_agent tool call
        create_calls = [
            call
            for call in result.get("tool_calls", [])
            if "create_agent" in call.get("tool", "").lower()
        ]

        if create_calls:
            tool_call = create_calls[0]
            args = tool_call.get("args", {})

            logger.info(f"Tool call args: {args}")

            # Verify LLM extracted correct name
            name = args.get("name", "")
            assert "CodeHelper" in str(
                name
            ), f"Expected name 'CodeHelper', got: {name}"

            # Verify LLM extracted correct description
            description = args.get("description", "")
            assert "Python" in str(
                description
            ), f"Expected description with 'Python', got: {description}"

        else:
            pytest.fail(
                f"Expected create_agent tool call, got: {result['tools']}"
            )



"""User data storage and retrieval tools for ToolRegistry.

This module provides tools for storing and retrieving user information
in the database. User data is stored in the User model and can be accessed
by key-value pairs.

Tools:
    - store_user_data: Store user information in the database
    - get_user_data: Retrieve user information from the database

Examples:
    >>> # Store user's name
    >>> store_user_data("name", "Alice")
    'Stored name: Alice'

    >>> # Retrieve user's name
    >>> get_user_data("name")
    'Alice'
"""

from airunner.components.llm.core.tool_registry import tool, ToolCategory
from airunner.components.user.data.user import User


@tool(
    name="store_user_data",
    category=ToolCategory.KNOWLEDGE,
    description="Store user information in the database by key-value pairs",
    requires_api=False,
    keywords=["save", "store", "user", "data", "remember", "profile", "setting"],
    input_examples=[
        {"key": "name", "value": "Alice"},
        {"key": "email", "value": "alice@example.com"},
        {"key": "timezone", "value": "America/New_York"},
    ],
)
def store_user_data(key: str, value: str) -> str:
    """Store user information in the database.

    This tool allows the LLM to remember information about the user by
    storing it in the User model. The data is persisted across sessions
    and can be retrieved later using get_user_data.

    Args:
        key: The data field name (e.g., "name", "email", "preferences")
        value: The data value to store


    Examples:
        >>> store_user_data("name", "Alice")
        'Stored name: Alice'

        >>> store_user_data("email", "alice@example.com")
        'Stored email: alice@example.com'

    Note:
        - Data is automatically saved to the database
        - Existing values will be overwritten
        - The key must be a valid field name in the User model
    """
    try:
        user = User.objects.get_or_create()
        setattr(user, key, value)
        user.save()
        return f"Stored {key}: {value}"
    except Exception as e:
        return f"Error storing data: {str(e)}"


@tool(
    name="get_user_data",
    category=ToolCategory.KNOWLEDGE,
    description="Retrieve user information from the database by key",
    requires_api=False,
    keywords=["get", "retrieve", "user", "data", "recall", "profile", "setting"],
    input_examples=[
        {"key": "name"},
        {"key": "email"},
        {"key": "preferences"},
    ],
)
def get_user_data(key: str) -> str:
    """Retrieve user information from the database.

    This tool allows the LLM to recall previously stored information
    about the user. The data must have been stored using store_user_data.

    Args:
        key: The data field name to retrieve


    Examples:
        >>> get_user_data("name")
        'Alice'

        >>> get_user_data("nonexistent")
        'No data found for key: nonexistent'

    Note:
        - Returns None if the key doesn't exist
        - Values are converted to strings
        - Automatically retrieves from current user
    """
    try:
        user = User.objects.get_or_create()
        value = getattr(user, key, None)
        if value is None:
            return f"No data found for key: {key}"
        return str(value)
    except Exception as e:
        return f"Error retrieving data: {str(e)}"
